#!/usr/bin/env python
# coding: utf-8

import requests

base_url = 'http://localhost:8080'

s = requests.Session()

# Sign in
s.get(base_url + '/signin?redirect=%2F')
s.post(base_url + '/signin', data={"userName": "root", "password": "root"})

# start ssh server
system_data = {
    "baseUrl": "http://localhost:8080",
    "information": "",
    "allowAccountRegistration": "false",
    "isCreateRepoOptionPublic": "true",
    "allowAnonymousAccess": "true",
    "activityLogLimit": "",
    "ssh": "on",
    "sshHost": "localhost",
    "sshPort": "29418",
}
s.post(base_url + '/admin/system', data=system_data)

# create test user
user_data = {
    "userName": "git-repo-user",
    "password": "user",
    "fullName": "git-repo-user",
    "mailAddress": "user@localhost"
}
s.post(base_url + '/admin/users/_newuser', data=user_data)

# create root repo
repo_data = {
    "owner": "root",
    "name": "repo",
    "description": "",
    "isPrivate": "false",
    "createReadme": "on"
}
s.post(base_url + '/new', data=repo_data)

# sign out
s.get(base_url + '/signout')

# sign in by user
s.get(base_url + '/signin?redirect=%2F')
s.post(base_url + '/signin', data={"userName": "git-repo-user", "password": "user"})

# create group
group_data ={
    "groupName": "group",
    "description": "",
    "memberName": "git-repo-user",
    "members": "git-repo-user:true"
}
s.post(base_url + '/groups/new', data=group_data)

# create Token
ret = s.post(base_url + '/git-repo-user/_personalToken', data={"note": "for test"})
html = ret.content.decode("utf-8")
find_text = "data-clipboard-text="
idx = html.find(find_text) + len(find_text) + 1
token = html[idx:idx + 40]

print("PRIVATE_KEY_GITBUCKET={}".format(token))




#!/usr/bin/env python

import os

import betamax
from betamax_serializers import pretty_json

record_mode = 'once'

from git_repo.services.service import RepositoryService
services = list(RepositoryService.service_map.keys())

if os.environ.get('TRAVIS_GH3'):
    # create default bogus values for tokens and namespaces if missing for pytest
    # to run without environment values
    # also if an environment variable is not set, then we don't want to record cassettes
    record_mode = 'never'
    for service in services:
        user_name = 'USERNAME_{}'.format(service.upper())
        token_name = 'PRIVATE_KEY_{}'.format(service.upper())
        namespace_name = '{}_NAMESPACE'.format(service.upper())
        if user_name not in os.environ:
            os.environ[user_name] = '_username_'.format(service)
        if token_name not in os.environ:
            os.environ[token_name] = '_token_{}_'.format(service)
        if namespace_name not in os.environ:
            os.environ[namespace_name] = '_namespace_{}_'.format(service)
else:
    # if running tests "locally" and not in travis, let's try to extract the keys from
    # the local configuration if there is some local configuration. And exposes them as
    # environment variables.
    import git, getpass
    config = git.config.GitConfigParser(os.path.join(os.environ['HOME'], '.gitconfig'))

    # handle the different forms of token configuration item (yup, technical debt bites here)
    get_section = lambda s: 'gitrepo "{}"'.format(s)
    get_username = lambda s: config.get_value(get_section(s), 'username',
                                    '_username_{}'.format(s)
                                )
    get_token = lambda s: config.get_value(get_section(s), 'token',
                            config.get_value(get_section(s), 'private_token',
                                config.get_value(get_section(s), 'privatekey',
                                        '_namespace_{}_'.format(s)
                                    )))

    for service in services:
        user_name = 'USERNAME_{}'.format(service.upper())
        token_name = 'PRIVATE_KEY_{}'.format(service.upper())
        namespace_name = '{}_NAMESPACE'.format(service.upper())
        if user_name not in os.environ:
            os.environ[user_name] = get_username(service)
        if token_name not in os.environ:
            os.environ[token_name] = get_token(service)
        if namespace_name not in os.environ:
            os.environ[namespace_name] = os.environ.get('GITREPO_NAMESPACE', '_namespace_{}_'.format(service))


def sanitize_token(interaction, current_cassette):
    # Exit early if the request did not return 200 OK because that's the
    # only time we want to look for Authorization-Token headers
    if interaction.data['response']['status']['code'] != 200:
        return

    headers = interaction.data['response']['headers']
    token = headers.get('Authorization')
    # If there was no token header in the response, exit
    if token is None:
        return

    # Otherwise, create a new placeholder so that when cassette is saved,
    # Betamax will replace the token with our placeholder.
    current_cassette.placeholders.append(
        cassette.Placeholder(placeholder='<AUTH_TOKEN>', replace=token)
    )

betamax.Betamax.register_serializer(pretty_json.PrettyJSONSerializer)

with betamax.Betamax.configure() as config:
    config.default_cassette_options['record_mode'] = record_mode
    config.cassette_library_dir = 'tests/integration/cassettes'
    config.default_cassette_options['serialize_with'] = 'prettyjson'
    #config.before_record(callback=sanitize_token)
    # generating placeholders in betamax configuration for each service's key and default namespace
    for service in services:
        config.define_cassette_placeholder('<PRIVATE_KEY_{}>'.format(service.upper()), os.environ.get('PRIVATE_KEY_{}'.format(service.upper())))
        config.define_cassette_placeholder('<{}_NAMESPACE>'.format(service.upper()), os.environ.get('{}_NAMESPACE'.format(service.upper())))




#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.kwargparse')

class KeywordArgumentParser:
    '''
    Argument parser tailored for use with docopt:

    Will parse all arguments returned by docopt, and will store them as instance
    attributes. Then it will launch actions based on the keyword argument list.

    * Parameter arguments (starting with -- or within <>) will automagically be
    parsed, dashes will be converted to underscores and stored as attributes of
    the parser instance.
    * Though if a parameter is matching a method decorated with "store_parameter",
    it won't be automagically setup, but will use that method instead.

    Then the keywords used as arguments of the program will be used to call an
    "action" method, which have been decorated with "register_action".

    '''
    _action_dict = dict()
    _parameter_dict = dict()

    def __init__(self, args):
        '''
        Stores the docopt args as class member
        '''
        self.args = args

    def init(self):
        '''
        method to be defined for anything that needs to be done before launching
        the parser.
        '''
        pass

    def run(self):
        '''
        This method iterates over the docopt's arguments and matches them against
        the parameters list. All leftover values are used to resolve the action to
        run, and it will run the action, or run the fallback() method if no action
        is found.
        '''
        self.init()

        args = []
        missed = []

        # go through setters
        for arg, value in self.args.items():
            if arg in self._parameter_dict:
                self._parameter_dict[arg](self, value)
                #log.debug('calling setter: {} → {}'.format(arg, value))
            elif arg.startswith('--') or arg.startswith('<'):
                arg_renamed = arg.lstrip('-<').rstrip('>').replace('-', '_')
                if not hasattr(self, arg_renamed):
                    setattr(self, arg_renamed, value)
                    #log.debug('auto-setting:   self.{} → {}'.format(arg_renamed, value))
            else:
                #log.debug('keeping: {} → {}'.format(arg, value))
                if self.args[arg]:
                    args.append(arg)

        if frozenset(args) in self._action_dict:
            #log.debug('running action: {}'.format(self._action_dict[frozenset(args)]))
            return self._action_dict[frozenset(args)](self)
        else:
            return self.fallback()

    def fallback(self):
        log.error('Unknown action.')
        log.error('Please consult help page (--help).')
        return 1



def store_parameter(parameter):
    '''
    Decorator for a parameter, use the full length parameter name as specified in the
    docopt configuration, like '--verbose'
    '''
    def decorator(fun):
        KeywordArgumentParser._parameter_dict[parameter] = fun
        return fun
    return decorator

def register_action(*args, **kwarg):
    '''
    Decorator for an action, the arguments order is not relevant, but it's best
    to use the same order as in the docopt for clarity.
    '''
    def decorator(fun):
        KeywordArgumentParser._action_dict[frozenset(args)] = fun
        return fun
    return decorator



#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.bitbucket')

from ..service import register_target, RepositoryService, ProgressBar
from ...exceptions import ResourceError, ResourceExistsError, ResourceNotFoundError
from ...tools import columnize

from pybitbucket.bitbucket import Client, Bitbucket
from pybitbucket.auth import BasicAuthenticator
from pybitbucket.pullrequest import PullRequest, PullRequestPayload
from pybitbucket.repository import (
        Repository, RepositoryPayload, RepositoryForkPayload,
        RepositoryForkPolicy, RepositoryType
)
from pybitbucket.snippet import Snippet, SnippetPayload
from pybitbucket.user import User

from requests import Request, Session
from requests.exceptions import HTTPError

from git.exc import GitCommandError

from lxml import html
import os, json, platform


@register_target('bb', 'bitbucket')
class BitbucketService(RepositoryService):
    fqdn = 'bitbucket.org'

    def __init__(self, *args, **kwarg):
        self.bb = Bitbucket(Client())
        super(BitbucketService, self).__init__(*args, **kwarg)

    def connect(self):
        if self._privatekey and ':' in self._privatekey:
            login, password = self._privatekey.split(':')
        else:
            login = self._username
            password = self._privatekey

        if not login or not password:
            raise ConnectionError('Could not connect to BitBucket. Please configure .gitconfig with your bitbucket credentials.')

        auth = BasicAuthenticator(login, password, 'z+git-repo+pub@m0g.net')
        self.bb.client.config = auth
        self.bb.client.session = self.bb.client.config.session = auth.start_http_session(self.bb.client.session)
        try:
            _ = self.bb.client.config.who_am_i()
        except ResourceError as err:
            raise ConnectionError('Could not connect to BitBucket. Not authorized, wrong credentials.') from err

    def create(self, user, repo, add=False):
        try:
            repo = Repository.create(
                    RepositoryPayload(dict(
                        fork_policy=RepositoryForkPolicy.ALLOW_FORKS,
                        is_private=False
                    )),
                    repository_name=repo,
                    owner=user,
                    client=self.bb.client
            )
            if add:
                self.add(user=user, repo=repo.name, tracking=self.name)
        except HTTPError as err:
            if '400' in err.args[0].split(' '):
                raise ResourceExistsError('Project {} already exists on this account.'.format(repo)) from err
            raise ResourceError("Couldn't complete creation: {}".format(err)) from err

    def fork(self, user, repo):
        # result = self.get_repository(user, repo).fork(
        #     RepositoryForkPayload(dict(name=repo)),
        #     owner=user)
        resp = self.bb.client.session.post(
            'https://api.bitbucket.org/1.0/repositories/{}/{}/fork'.format(user, repo),
            data={'name': repo}
        )
        if 404 == resp.status_code:
            raise ResourceNotFoundError("Couldn't complete fork: {}".format(resp.content.decode('utf-8')))
        elif 200 != resp.status_code:
            raise ResourceError("Couldn't complete fork: {}".format(resp.content.decode('utf-8')))
        result = resp.json()
        return '/'.join([result['owner'], result['slug']])

    def delete(self, repo, user=None):
        if not user:
            user = self.user
        try:
            self.get_repository(user, repo).delete()
        except HTTPError as err:
            if '404' in err.args[0].split(' '):
                raise ResourceNotFoundError("Cannot delete: repository {}/{} does not exists.".format(user, repo)) from err
            raise ResourceError("Couldn't complete deletion: {}".format(err)) from err

    def list(self, user, _long=False):
        try:
            user = User.find_user_by_username(user)
        except HTTPError as err:
            raise ResourceNotFoundError("User {} does not exists.".format(user)) from err

        repositories = user.repositories()
        if not _long:
            yield "{}"
            repositories = list(repositories)
            yield ("Total repositories: {}".format(len(repositories)),)
            yield from columnize(["/".join([user.username, repo.name]) for repo in repositories])
        else:
            yield "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}"
            yield ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif', 'Name']
            for repo in repositories:
                # if repo.updated_at.year < datetime.now().year:
                #     date_fmt = "%b %d %Y"
                # else:
                #     date_fmt = "%b %d %H:%M"

                status = ''.join([
                    'F' if getattr(repo, 'parent', None) else ' ', # is a fork?
                    'P' if repo.is_private else ' ',               # is private?
                ])
                yield [
                    # status
                    status,
                    # stats
                    str(len(list(repo.commits()))),       # number of commits
                    str(len(list(repo.pullrequests()))),  # number of pulls
                    str('N.A.'),                          # number of issues
                    str(len(list(repo.forks()))),         # number of forks
                    str('N.A.'),                          # number of contributors
                    str(len(list(repo.watchers()))),      # number of subscribers
                    str('N.A.'),                          # number of ♥
                    # info
                    repo.language or '?',                 # language
                    repo.updated_on,                      # date
                    '/'.join([user.username, repo.name]), # name
                ]

    def _format_gist(self, gist):
        return gist.split('/')[-1] if gist.startswith('http') else gist

    def gist_list(self, gist=None):
        if not gist:
            for snippet in list(self.bb.snippetByOwner(owner=self.user)):
                if isinstance(snippet, Snippet):
                    yield (snippet.links['html']['href'], snippet.title)
        else:
            try:
                snippet = next(self.bb.snippetByOwnerAndSnippetId(owner=self.user, snippet_id=self._format_gist(gist)))
                for snippet_file in snippet.filenames:
                    yield ('N.A.',
                            0,
                            snippet_file)
            except HTTPError as err:
                if '404' in err.args[0].split(' '):
                    raise ResourceNotFoundError("Could not find snippet {}.".format(gist)) from err
                raise ResourceError("Couldn't fetch snippet details: {}".format(err)) from err

    def gist_fetch(self, gist, fname=None):
        gist = self._format_gist(gist)
        try:
            user = self.user
            snippet = next(self.bb.snippetByOwnerAndSnippetId(owner=user, snippet_id=gist))
        except HTTPError as err:
            if '404' in err.args[0].split(' '):
                raise ResourceNotFoundError("Could not find snippet {}.".format(gist)) from err
            raise ResourceError("Couldn't fetch snippet details: {}".format(err)) from err
        if len(snippet.filenames) == 1 and not fname:
            gist_file = snippet.filenames[0]
        else:
            if fname in snippet.filenames:
                gist_file = fname
            else:
                raise ResourceNotFoundError('Could not find file within gist.')

        return self.bb.client.session.get(
                    'https://bitbucket.org/!api/2.0/snippets/{}/{}/files/{}'.format(user, gist, gist_file)
                ).content.decode('utf-8')

    def gist_clone(self, gist):
        gist = self._format_gist(gist)
        try:
            snippet = next(self.bb.snippetByOwnerAndSnippetId(owner=self.user, snippet_id=gist))
            remotes = {it['name']: it['href'] for it in snippet.links['clone']}
            if 'ssh' in remotes:
                remote = remotes['ssh']
            elif 'ssh' in remotes:
                remote = remotes['https']
            else:
                raise ResourceError("Couldn't find appropriate method for cloning.")
        except HTTPError as err:
            if '404' in err.args[0].split(' '):
                raise ResourceNotFoundError("Could not find snippet {}.".format(gist)) from err
            raise ResourceError("Couldn't fetch snippet details: {}".format(err)) from err
        remote = self.repository.create_remote('gist', remote)
        self.pull(remote, 'master')

    def gist_create(self, gist_pathes, description, secret=False):
        def load_file(fname, path='.'):
            return open(os.path.join(path, fname), 'r')

        gist_files = dict()
        for gist_path in gist_pathes:
            if not os.path.isdir(gist_path):
                gist_files[os.path.basename(gist_path)] = load_file(gist_path)
            else:
                for gist_file in os.listdir(gist_path):
                    if not os.path.isdir(os.path.join(gist_path, gist_file)) and not gist_file.startswith('.'):
                        gist_files[gist_file] = load_file(gist_file, gist_path)

        try:
            snippet = Snippet.create(
                files=gist_files,
                payload=SnippetPayload(
                    payload=dict(
                        title=description,
                        scm=RepositoryType.GIT,
                        is_private=secret
                    )
                ),
                client=self.bb.client
            )

            return snippet.links['html']['href']
        except HTTPError as err:
            raise ResourceError("Couldn't create snippet: {}".format(err)) from err

    def gist_delete(self, gist_id):
        try:
            snippet = next(self.bb.snippetByOwnerAndSnippetId(owner=self.user, snippet_id=gist_id)).delete()
        except HTTPError as err:
            if '404' in err.args[0].split(' '):
                raise ResourceNotFoundError("Could not find snippet {}.".format(gist_id)) from err
            raise ResourceError("Couldn't delete snippet: {}".format(err)) from err

    def request_create(self, onto_user, onto_repo, from_branch, onto_branch, title=None, description=None, auto_slug=False, edit=None):
        try:
            onto_project = self.get_repository(onto_user, onto_repo)

            from_reposlug = self.guess_repo_slug(self.repository, self, resolve_targets=['{service}'])
            if from_reposlug:
                from_user, from_repo = from_reposlug.split('/')
                if (onto_user, onto_repo) == (from_user, from_repo):
                    from_project = onto_project
                else:
                    from_project = self.get_repository(from_user, from_repo)
            else:
                from_project = None

            # when no repo slug has been given to `git-repo X request create`
            # then chances are current project is a fork of the target
            # project we want to push to
            if auto_slug and onto_project.fork:
                onto_user = onto_project.parent.owner.login
                onto_repo = onto_project.parent.name
                onto_project = self.repository(onto_user, onto_repo)

            # if no onto branch has been defined, take the default one
            # with a fallback on master
            if not from_branch:
                from_branch = self.repository.active_branch.name

            # if no from branch has been defined, chances are we want to push
            # the branch we're currently working on
            if not onto_branch:
                onto_branch = self.get_project_default_branch(onto_project)

            from_target = '{}:{}'.format(from_user, from_branch)
            onto_target = '{}/{}:{}'.format(onto_user, onto_project, onto_branch)

            # translate from github username to git remote name
            if not title and not description and edit:
                title, description = edit(self.repository, from_branch, onto_target)
                if not title and not description:
                    raise ArgumentError('Missing message for request creation')

            request = PullRequest.create(
                        PullRequestPayload(
                            payload=dict(
                                title=title,
                                description=description or '',
                                destination=dict(
                                    branch=dict(name=onto_branch)
                                ),
                                source=dict(
                                    repository=dict(full_name='/'.join([from_user, from_repo])),
                                    branch=dict(name=from_branch)
                                )
                            )
                        ),
                        repository_name=onto_repo,
                        owner=onto_user,
                        client=self.bb.client
                    )

            return {
                'local': from_branch,
                'remote': onto_branch,
                'ref': request.id,
                'project': '/'.join([onto_user, onto_repo]),
                'url': request.links['html']['href']
            }

        except HTTPError as err:
            status_code = hasattr(err, 'code') and err.code or err.response.status_code
            if 404 == status_code:
                raise ResourceNotFoundError("Couldn't create request, project not found: {}".format(onto_repo)) from err
            elif 400 == status_code and 'branch not found' in err.format_message():
                raise ResourceNotFoundError("Couldn't create request, branch not found: {}".format(from_branch)) from err
            raise ResourceError("Couldn't create request: {}".format(err)) from err

    def request_list(self, user, repo):
        requests = set(
            (
                str(r.id),
                r.title,
                r.links['html']['href']
            ) for r in self.bb.repositoryPullRequestsInState(
                owner=user,
                repository_name=repo,
                state='open'
            ) if not isinstance(r, dict) # if no PR is empty, result is a dict
        )
        for pull in sorted(requests):
            try:
                yield pull
            except Exception as err:
                log.warn('Error while fetching request information: {}'.format(pull))

    def request_fetch(self, user, repo, request, pull=False):
        if pull:
            raise NotImplementedError('Pull operation on requests for merge are not yet supported')

        pb = ProgressBar()
        pb.setup(self.name)

        try:
            local_branch_name = 'requests/bitbucket/{}'.format(request)
            pr = next(self.bb.repositoryPullRequestByPullRequestId(
                owner=user,
                repository_name=repo,
                pullrequest_id=request
            ))
            source_branch = pr.source['branch']['name']
            source_slug = pr.source['repository']['full_name']
            source_url = pr.source['repository']['links']['html']['href']
            remote_name = 'requests/bitbucket/{}'.format(source_slug).replace('/', '-')
            try:
                remote = self.repository.remote(name=remote_name)
            except ValueError:
                remote = self.repository.create_remote(name=remote_name, url=source_url)
            refspec = '{}:{}'.format(source_branch, local_branch_name)
            refs = remote.fetch(refspec, progress=pb)
            for branch in self.repository.branches:
                if branch.name == local_branch_name:
                    branch.set_tracking_branch(remote.refs[0])
            return local_branch_name
        except HTTPError as err:
            if '404' in err.args[0].split(' '):
                raise ResourceNotFoundError('Could not find opened request #{}'.format(request)) from err
            raise ResourceError("Couldn't delete snippet: {}".format(err)) from err
        except GitCommandError as err:
            if 'Error when fetching: fatal: ' in err.command[0]:
                raise ResourceNotFoundError('Could not find opened request #{}'.format(request)) from err
            raise err

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        session = Session()

        key_name = 'git-repo@{}'.format(platform.node())

        # get login page
        log.info('» Login to bitbucket…')

        login_url = "https://bitbucket.org/account/signin/?next=/".format(login)

        session.headers.update({
            "User-Agent":"Mozilla/5.0 (X11; Linux x86_64) " \
            "AppleWebKit/537.36 (KHTML, like Gecko) " \
            "Chrome/52.0.2743.82 Safari/537.36"
        })

        result = session.get(login_url)
        tree = html.fromstring(result.text)

        # extract CSRF token

        authenticity_token = list(set(tree.xpath("//input[@name='csrfmiddlewaretoken']/@value")))[0]

        # do login

        payload = {
            'username': login,
            'password': password,
            'csrfmiddlewaretoken': authenticity_token
        }

        result = session.post(
            login_url,
            data = payload,
            headers = dict(referer=login_url)
        )
        tree = html.fromstring(result.text)

        # extract username

        try:
            username = json.loads(tree.xpath('//meta/@data-current-user')[0])['username']
        except KeyError:
            raise ResourceNotFoundError('Invalid login. Please make sure you\'re using your bitbucket email address as username!')

        app_password_url ='https://bitbucket.org/account/user/{}/app-passwords/new'.format(username)

        # load app password page
        log.info('» Generating app password…')

        result = session.get(
            app_password_url,
            headers=dict(referer=app_password_url)
        )
        tree = html.fromstring(result.content)

        if 'git-repo@{}'.format(platform.node()) in result:
            log.warn("A duplicate key is being created!")

        # generate app password
        log.info('» App password is setup with following scopes:')
        log.info('»     account, team, project:write, repository:admin, repository:delete')
        log.info('»     pullrequest:write, snippet, snippet:write')

        authenticity_token = list(set(tree.xpath("//input[@name='csrfmiddlewaretoken']/@value")))[0]

        payload = dict(
            name=key_name,
            scope=['account',
                    'team',
                    'project',
                    'project:write',
                    'repository',
                    'pullrequest:write',
                    'repository:admin',
                    'repository:delete',
                    'snippet',
                    'snippet:write'
                    ],
            csrfmiddlewaretoken=authenticity_token
        )
        result = session.post(
            app_password_url,
            data=payload,
            headers=dict(referer=app_password_url)
        )
        tree = html.fromstring(result.content)

        password = json.loads(tree.xpath('//section/@data-app-password')[0])['password']

        return password, username

    def get_parent_project_url(self, user, project, rw=True):
        project = self.get_repository(user, project)
        if not project or not hasattr(project, 'parent') or not project.parent:
            return None
        return self.format_path(
            repository=project.parent.name,
            namespace=project.parent.owner.login,
            rw=True)

    @property
    def user(self):
        try:
            user = next(self.bb.userForMyself()).username
            return user
        except (HTTPError, AttributeError) as err:
            raise ResourceError("Couldn't find the current user: {}".format(err)) from err


    def get_repository(self, user, repo):
        try:
            return next(self.bb.repositoryByOwnerAndRepositoryName(owner=user, repository_name=repo))
        except HTTPError as err:
            raise ResourceNotFoundError('Cannot retrieve repository: {}/{} does not exists.'.format(user, repo))

    @staticmethod
    def is_repository_empty(project):
        return project.size == 0

    @staticmethod
    def get_project_default_branch(project):
        return project.mainbranch.get('name', 'master')



#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.gitlab')

from ..service import register_target, RepositoryService
from ...exceptions import ArgumentError, ResourceError, ResourceExistsError, ResourceNotFoundError
from ...tools import columnize

import gitlab
from gitlab.exceptions import GitlabListError, GitlabCreateError, GitlabGetError

from git.exc import GitCommandError

import os
import json, time
import dateutil.parser
from datetime import datetime

@register_target('lab', 'gitlab')
class GitlabService(RepositoryService):
    fqdn = 'gitlab.com'

    def __init__(self, *args, **kwarg):
        self.gl = gitlab.Gitlab(self.url_ro)
        super().__init__(*args, **kwarg)

    def connect(self):
        self.gl.ssl_verify = self.session_certificate or not self.session_insecure
        if self.session_proxy:
            self.gl.session.proxies.update(self.session_proxy)

        self.gl.set_url(self.url_ro)
        self.gl.set_token(self._privatekey)
        self.gl.token_auth()
        self.username = self.gl.user.username

    def create(self, user, repo, add=False):
        try:
            group = self.gl.groups.search(user)
            data = {'name': repo}
            if group:
                data['namespace_id'] = group[0].id
            self.gl.projects.create(data=data)
        except GitlabCreateError as err:
            if json.loads(err.response_body.decode('utf-8'))['message']['name'][0] == 'has already been taken':
                raise ResourceExistsError("Project already exists.") from err
            else:
                raise ResourceError("Unhandled error.") from err
        if add:
            self.add(user=user, repo=repo, tracking=self.name)

    def fork(self, user, repo):
        try:
            return self.gl.projects.get('{}/{}'.format(user, repo)).forks.create({}).path_with_namespace
        except GitlabCreateError as err:
            if json.loads(err.response_body.decode('utf-8'))['message']['name'][0] == 'has already been taken':
                raise ResourceExistsError("Project already exists.") from err
            else:
                raise ResourceError("Unhandled error: {}".format(err)) from err

    def delete(self, repo, user=None):
        if not user:
            user = self.user
        try:
            repository = self.gl.projects.get('{}/{}'.format(user, repo))
            if repository:
                result = self.gl.delete(repository.__class__, repository.id)
            if not repository or not result:
                raise ResourceNotFoundError("Cannot delete: repository {}/{} does not exists.".format(user, repo))
        except GitlabGetError as err:
            if err.response_code == 404:
                raise ResourceNotFoundError("Cannot delete: repository {}/{} does not exists.".format(user, repo)) from err
            elif err.response_code == 403:
                raise ResourcePermissionError("You don't have enough permissions for deleting the repository. Check the namespace or the private token's privileges") from err
        except Exception as err:
            raise ResourceError("Unhandled exception: {}".format(err)) from err

    def list(self, user, _long=False):
        if not self.gl.users.search(user):
            raise ResourceNotFoundError("User {} does not exists.".format(user))

        repositories = self.gl.projects.list(author=user)
        if not _long:
            repositories = list([repo.path_with_namespace for repo in repositories])
            yield "{}"
            yield ("Total repositories: {}".format(len(repositories)),)
            yield from columnize(repositories)
        else:
            yield "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}"
            yield ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif\t', 'Name']
            for repo in repositories:
                time.sleep(0.5)
                repo_last_activity_at = dateutil.parser.parse(repo.last_activity_at)
                if repo_last_activity_at.year < datetime.now().year:
                    date_fmt = "%b %d %Y"
                else:
                    date_fmt = "%b %d %H:%M"

                status = ''.join([
                    'F' if False else ' ',               # is a fork?
                    'P' if repo.visibility_level == 0 else ' ',            # is private?
                ])
                yield [
                                                               # status
                    status,
                                                               # stats
                    str(len(list(repo.commits.list()))),       # number of commits
                    str(len(list(repo.mergerequests.list()))), # number of pulls
                    str(len(list(repo.issues.list()))),        # number of issues
                    str(repo.forks_count),                     # number of forks
                    str(len(list(repo.members.list()))),       # number of contributors
                    'N.A.',                                    # number of subscribers
                    str(repo.star_count),                      # number of ♥
                                                               # info
                    'N.A.',                                    # language
                    repo_last_activity_at.strftime(date_fmt),  # date
                    repo.name_with_namespace,                  # name
                ]

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        gl = gitlab.Gitlab(url='https://{}'.format(cls.fqdn), email=login, password=password)
        gl.auth()
        return gl.user.private_token

    def _deconstruct_snippet_uri(self, uri):
        path = uri.split('https://{}/'.format(self.fqdn))[-1].split('/')
        if 4 == len(path):
            user, project_name, _, snippet_id = path
        elif 2 == len(path):
            _, snippet_id = path
            project_name = None
            user = None
        elif 1 == len(path):
            snippet_id = path[0]
            project_name = None
            user = None
        else:
            raise ResourceNotFoundError('URL is not of a snippet')
        return (user, project_name, snippet_id)

    def gist_list(self, project=None):
        yield "{:45.45} {}"
        yield 'title', 'url'
        if not project:
            try:
                for snippet in self.gl.snippets.list():
                    yield snippet.title, snippet.web_url
            except GitlabListError as err:
                if err.response_code == 404:
                    raise ResourceNotFoundError('Feature not available, please upgrade your gitlab instance.') from err
                raise ResourceError('Cannot list snippet') from err
        else:
            if '/' not in project:
                project = '/'.join([self.username, project])
            try:
                project = self.gl.projects.get(project)
                for snippet in project.snippets.list():
                    yield (snippet.web_url, snippet.title)
            except GitlabGetError as err:
                raise ResourceNotFoundError('Could not retrieve project "{}".'.format(project)) from err

    def gist_fetch(self, snippet, fname=None):
        if fname:
            raise ArgumentError('Snippets contain only single file in gitlab.')
        try:
            *_, snippet_id = self._deconstruct_snippet_uri(snippet)
            snippet = self.gl.snippets.get(id=snippet_id)
        except GitlabGetError as err:
            if err.response_code == 404:
                if "The page you're looking for could not be found." in err.response_body.decode('utf-8'):
                    raise ResourceNotFoundError('Feature not available, please upgrade your gitlab instance.') from err
                raise ResourceNotFoundError('Cannot fetch snippet') from err
            raise ResourceError('Cannot fetch snippet') from err
        except Exception as err:
            raise ResourceNotFoundError('Could not find snippet') from err

        return snippet.raw().decode('utf-8')

    def gist_clone(self, gist):
        raise ArgumentError('Snippets cannot be cloned in gitlab.')

    def gist_create(self, gist_pathes, description, secret=False):
        def load_file(fname, path='.'):
            with open(os.path.join(path, fname), 'r') as f:
                return f.read()

        if len(gist_pathes) > 2:
            raise ArgumentError('Snippets contain only single file in gitlab.')

        data = {
            'title': description,
            'visibility_level': 0 if secret else 20
        }

        try:

            if len(gist_pathes) == 2:
                project = gist_pathes[0]
                if '/' in project:
                    *namespace, project = project.split('/')
                    namespace = '/'.join(namespace)
                else:
                    namespace = self.username
                gist_path = gist_pathes[1]
                data.update({
                        'project_id': '/'.join([namespace, project]),
                        'code': load_file(gist_path),
                        'file_name': os.path.basename(gist_path),
                    }
                )
                gist = self.gl.project_snippets.create(data)

            elif len(gist_pathes) == 1:
                gist_path = gist_pathes[0]
                data.update({
                        'content': load_file(gist_path),
                        'file_name': os.path.basename(gist_path),
                    }
                )
                gist = self.gl.snippets.create(data)

            return gist.web_url
        except GitlabCreateError as err:
            if err.response_code == 422:
                raise ResourceNotFoundError('Feature not available, please upgrade your gitlab instance.') from err
            raise ResourceError('Cannot create snippet') from err

    def gist_delete(self, snippet):
        try:
            _, project, snippet_id = self._deconstruct_snippet_uri(snippet)
            if project:
                if '/' in project:
                    *namespace, project = project.split('/')
                    namespace = '/'.join(namespace)
                else:
                    namespace = self.username
                snippet = self.gl.projects.get(
                        '/'.join([namespace, project])
                    ).snippets.get(id=snippet_id)
            else:
                snippet = self.gl.snippets.get(id=snippet_id)
        except GitlabCreateError as err:
            if err.response_code == 422:
                raise ResourceNotFoundError('Cannot delete snippet, please upgrade your gitlab instance.') from err
            raise ResourceError('Cannot delete snippet') from err
        except Exception as err:
            raise ResourceNotFoundError('Could not find snippet') from err

        return snippet.delete()

    def request_create(self, onto_user, onto_repo, from_branch, onto_branch, title=None, description=None, auto_slug=False, edit=None):
        try:
            onto_project = self.gl.projects.get('/'.join([onto_user, onto_repo]))

            if not onto_project:
                raise ResourceNotFoundError('Could not find project `{}/{}`!'.format(onto_user, onto_repo))

            from_reposlug = self.guess_repo_slug(self.repository, self, resolve_targets=['{service}'])
            if from_reposlug:
                from_user, from_repo = from_reposlug.split('/')
                if (onto_user, onto_repo) == (from_user, from_repo):
                    from_project = onto_project
                else:
                    from_project = self.gl.projects.get('/'.join([from_user, from_repo]))
            else:
                from_project = None

            if not from_project:
                raise ResourceNotFoundError('Could not find project `{}/{}`!'.format(from_user, from_repo))

            # when no repo slug has been given to `git-repo X request create`
            # then chances are current project is a fork of the target
            # project we want to push to
            if auto_slug and 'forked_from_project' in onto_project.as_dict():
                parent = self.gl.projects.get(onto_project.forked_from_project['id'])
                onto_user, onto_repo = parent.namespace.path, parent.path
                onto_project = self.gl.projects.get('/'.join([onto_user, onto_repo]))

            # if no onto branch has been defined, take the default one
            # with a fallback on master
            if not from_branch:
                from_branch = self.repository.active_branch.name
            # if no from branch has been defined, chances are we want to push
            # the branch we're currently working on
            if not onto_branch:
                onto_branch = self.get_project_default_branch(onto_project)

            onto_target = '{}:{}'.format(onto_user, onto_project.path_with_namespace, onto_branch)

            # translate from gitlab username to git remote name
            if not title and not description and edit:
                title, description = edit(self.repository, from_branch, onto_target)
                if not title and not description:
                    raise ArgumentError('Missing message for request creation')

            request = self.gl.project_mergerequests.create(
                    project_id=from_project.id,
                    data={
                        'source_branch': from_branch,
                        'target_branch': onto_branch,
                        'target_project_id': onto_project.id,
                        'title': title,
                        'description': description
                    }
            )

            return {
                'local': from_branch,
                'project': '/'.join([onto_user, onto_repo]),
                'remote': onto_branch,
                'url': request.web_url,
                'ref': request.iid
            }

        except GitlabGetError as err:
            raise ResourceNotFoundError(err) from err
        except Exception as err:
            raise ResourceError("Unhandled error: {}".format(err)) from err

    def request_list(self, user, repo):
        project = self.gl.projects.get('/'.join([user, repo]))
        yield "{:>3}\t{:<60}\t{:2}"
        yield ('id', 'title', 'URL')
        for mr in self.gl.project_mergerequests.list(project_id=project.id):
            yield ( str(mr.iid),
                    mr.title,
                    mr.web_url
                    )

    def request_fetch(self, user, repo, request, pull=False, force=False):
        if pull:
            raise NotImplementedError('Pull operation on requests for merge are not yet supported')
        try:
            for remote in self.repository.remotes:
                if remote.name == self.name:
                    local_branch_name = 'requests/gitlab/{}'.format(request)
                    self.fetch(
                        remote,
                       'merge-requests/{}/head'.format(request),
                        local_branch_name,
                        force
                    )
                    return local_branch_name
            else:
                raise ResourceNotFoundError('Could not find remote {}'.format(self.name))
        except GitCommandError as err:
            if 'Error when fetching: fatal: Couldn\'t find remote ref' in err.command[0]:
                raise ResourceNotFoundError('Could not find opened request #{}'.format(request)) from err
            raise err

    def get_parent_project_url(self, user, repo, rw=True):
        project = self.gl.projects.get('/'.join([user, repo]))
        parent = None
        if hasattr(project, 'forked_from_project'):
            parent = self.gl.projects.get(project.forked_from_project['id'])
        if not parent:
            return None
        return self.format_path(
                repository=parent.path,
                namespace=parent.namespace.path,
                rw=rw)

    @property
    def user(self):
        return self.gl.user.username

    def get_repository(self, user, repo):
        try:
            return self.gl.projects.get('{}/{}'.format(user, repo))
        except GitlabGetError as err:
            if err.response_code == 404:
                raise ResourceNotFoundError("Cannot get: repository {}/{} does not exists.".format(user, repo)) from err

    @staticmethod
    def get_project_default_branch(project):
        return project.default_branch or 'master'

    @staticmethod
    def is_repository_empty(project):
        try:
            project.repository_tree()
            return True
        except gitlab.exceptions.GitlabGetError:
            return False




#!/usr/bin/env python

import logging

#################################################################################
# Enable logging

log = logging.getLogger('test.gogs')

#################################################################################

import os
import sys
import pytest
from urllib.parse import urlparse

from tests.helpers import GitRepoTestCase

from git_repo.services.service import gogs
from git_repo.exceptions import ResourceExistsError

from betamax import Betamax

GOGS_URL=os.environ.get('GOGS_URL', 'https://try.gogs.io')

with Betamax.configure() as config:
    config.define_cassette_placeholder('<GOGS_URL>', GOGS_URL)

class Test_Gogs(GitRepoTestCase):
    log = log

    @property
    def local_namespace(self):
        if 'GOGS_NAMESPACE' in os.environ:
            return os.environ['GOGS_NAMESPACE']
        return 'git-repo-test'

    def get_service(self):
        url = urlparse(GOGS_URL)
        return gogs.GogsService(c={
            '__name__': 'gitrepo "gogs"',
            'fqdn': url.hostname,
            'port': url.port,
            'scheme': url.scheme,
            'insecure': 'yes',
            })

    def get_requests_session(self):
        return self.service.gg.session

    def test_00_fork(self):
        pass
        #self.action_fork(local_namespace=self.local_namespace,
        #                 remote_namespace='sigmavirus24',
        #                 repository='github3-py')

    def test_01_create__new(self):
        self.action_create(namespace=self.local_namespace,
                           repository='foobar')

    def test_01_create__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace=self.local_namespace,
                            repository='git-repo')

    def test_01_create_group__new(self):
        self.action_create(namespace='guyzmo',
                           repository='foobar')

    def test_01_create_group__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace='guyzmo',
                            repository='git-repo')


    def test_02_delete(self):
        self.action_delete(namespace='guyzmo',
                           repository='foobar')

    def test_03_delete_nouser(self):
        self.action_delete(repository='foobar')

    def test_04_clone(self):
        self.action_clone(namespace=self.local_namespace,
                          repository='git-repo')

    def test_05_add(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo')

    def test_06_add__name(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        name='test0r')

    def test_07_add__alone(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        alone=True)

    def test_08_add__alone_name(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        name='test0r',
                        alone=True)

    def test_09_add__default(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        tracking='gogs')

    def test_10_add__default_name(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        name='test0r',
                        tracking='gogs')

    def test_11_add__alone_default(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        alone=True,
                        tracking='gogs')

    def test_12_add__alone_default_name(self):
        self.action_add(namespace=self.local_namespace,
                        repository='git-repo',
                        alone=True,
                        name='test0r',
                        tracking='gogs')

    def test_13_open(self):
        self.action_open(namespace=self.local_namespace,
                         repository='git-repo')





#!/usr/bin/env python
import sys
import logging
log = logging.getLogger('git_repo.gogs')

from ..service import register_target, RepositoryService, os
from ...exceptions import ResourceError, ResourceExistsError, ResourceNotFoundError
from ...tools import columnize

from gogs_client import GogsApi, GogsRepo, Token, UsernamePassword, ApiFailure
from requests import Session, HTTPError
from urllib.parse import urlparse, urlunparse
from datetime import datetime
import dateutil.parser
import functools

from git import config as git_config
from git.exc import GitCommandError

class GogsClient(GogsApi):
    def __init__(self):
        self.session = Session()

    def setup(self, *args, **kwarg):
        super().__init__(*args, session=self.session, **kwarg)

    def set_token(self, token):
        self.auth = Token(token)

    def set_default_private(self, p):
        self.default_private = p

    def setup_session(self, ssl_config, proxy=dict()):
        self.session.verify = ssl_config
        self.session.proxies.update(proxy)
        # this is required to detect fresh GoGS server that redirects everything to /install
        self.session.max_redirects = 0

    @property
    def username(self):
        if not hasattr(self, '_username'):
            self._username = self.authenticated_user(self.auth).username
        return self._username

    def orgs(self):
        orgs = self._check_ok(self._get('/user/orgs', auth=self.auth)).json()
        #return [gogs_client.GogsUser.from_json(org) for org in orgs]
        return [org['username'] for org in orgs]

    def create_repository(self, user, repo):
        if user == self.username:
            repository = self.create_repo(self.auth, name=repo, private=self.default_private)
        elif user in self.orgs():
            data = dict(name=repo, private=self.default_private)
            response = self._post('/org/{}/repos'.format(user), auth=self.auth, data=data)
            repository = GogsRepo.from_json(self._check_ok(response).json())
        else:
            data = dict(name=repo, private=self.default_private)
            response = self._post('/admin/users/{}/repos'.format(user), auth=self.auth, data=data)
            repository = GogsRepo.from_json(self._check_ok(response).json())

    def delete_repository(self, user, repo):
        return self.delete_repo(self.auth, user, repo)

    def repository(self, user, repo):
        return self.get_repo(self.auth, user, repo)

    def repositories(self, user):
        r = self._get('/user/repos', auth=self.auth)
        repositories = self._check_ok(r).json()
        repositories = [repo for repo in repositories if repo['owner']['username'] == user]
        return repositories

@register_target('gg', 'gogs')
class GogsService(RepositoryService):
    fqdn = 'try.gogs.io'

    def __init__(self, *args, **kwargs):
        self.gg = GogsClient()

        super().__init__(*args, **kwargs)

    def connect(self):
        self.gg.setup(self.url_ro)
        self.gg.set_token(self._privatekey)
        self.gg.set_default_private(self.default_create_private)
        self.gg.setup_session(
                self.session_certificate or not self.session_insecure,
                self.session_proxy)
        try:
            self.username = self.user  # Call to self.gg.authenticated_user()
        except HTTPError as err:
            if err.response is not None and err.response.status_code == 401:
                if not self._privatekey:
                    raise ConnectionError('Could not connect to GoGS. '
                                          'Please configure .gitconfig '
                                          'with your gogs private key.') from err
                else:
                    raise ConnectionError('Could not connect to GoGS. '
                                          'Check your configuration and try again.') from err
            else:
                raise err

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        import platform
        name = 'git-repo token used on {}'.format(platform.node())
        gg = GogsApi(cls.build_url(cls))
        auth = UsernamePassword(login, password)
        tokens = gg.get_tokens(auth, login)
        tokens = dict((token.name, token.token) for token in tokens)
        if name in tokens:
            return tokens[name]
        if 'git-repo token' in tokens:
            return tokens['git-repo token']
        token = gg.create_token(auth, name, login)
        return token.token

    @property
    def user(self):
        return self.gg.username

    def create(self, user, repo, add=False):
        try:
            self.gg.create_repository(user, repo)
        except ApiFailure as err:
            if err.status_code == 422:
                raise ResourceExistsError("Project already exists.") from err
            else:
                raise ResourceError("Unhandled error.") from err
        except Exception as err:
            raise ResourceError("Unhandled exception: {}".format(err)) from err
        if add:
            self.add(user=self.username, repo=repo, tracking=self.name)

    def fork(self, user, repo):
        raise NotImplementedError

    def delete(self, repo, user=None):
        if not user:
            user = self.username
        try:
            self.gg.delete_repository(user, repo)
        except ApiFailure as err:
            if err.status_code == 404:
                raise ResourceNotFoundError("Cannot delete: repository {}/{} does not exists.".format(user, repo)) from err
            elif err.status_code == 403:
                raise ResourcePermissionError("You don't have enough permissions for deleting the repository. Check the namespace or the private token's privileges") from err
            elif err.status_code == 422:
                raise ResourceNotFoundError("Cannot delete repository {}/{}: user {} does not exists.".format(user, repo, user)) from err
            raise ResourceError("Unhandled error: {}".format(err)) from err
        except Exception as err:
            raise ResourceError("Unhandled exception: {}".format(err)) from err

    def list(self, user, _long=False):
        repositories = self.gg.repositories(user)
        if user != self.username and not repositories and user not in self.orgs:
            raise ResourceNotFoundError("Unable to list namespace {} - only authenticated user and orgs available for listing.".format(user))
        if not _long:
            repositories = list([repo['full_name'] for repo in repositories])
            yield "{}"
            yield ("Total repositories: {}".format(len(repositories)),)
            yield from columnize(repositories)
        else:
            yield "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}"
            yield ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif\t', 'Name']
            for repo in repositories:
                repo_updated_at = dateutil.parser.parse(repo['updated_at'])
                if repo_updated_at.year < datetime.now().year:
                    date_fmt = "%b %d %Y"
                else:
                    date_fmt = "%b %d %H:%M"
                status = ''.join([
                    'F' if repo['fork'] else ' ',          # is a fork?
                    'P' if repo['private'] else ' ',       # is private?
                ])
                try:
                    issues = self.gg._check_ok(self.gg._get('/repos/{}/issues'.format(repo['full_name']), auth=self.auth)).json()
                except Exception:
                    issues = []
                yield [
                    # status
                    status,
                    # stats
                    str(len(list(()))),                    # number of commits
                    str(len(list(()))),                    # number of pulls
                    str(len(list(issues))),                # number of issues
                    str(repo.get('forks_count') or 0),     # number of forks
                    str(len(list(()))),                    # number of contributors
                    str(repo.get('watchers_count') or 0),  # number of subscribers
                    str(repo.get('stars_count') or 0),     # number of ♥
                    # info
                    repo.get('language') or '?',           # language
                    repo_updated_at.strftime(date_fmt),    # date
                    repo['full_name'],                     # name
                ]

    def gist_list(self, gist=None):
        raise NotImplementedError

    def gist_fetch(self, gist, fname=None):
        raise NotImplementedError

    def gist_clone(self, gist):
        raise NotImplementedError

    def gist_create(self, gist_pathes, description, secret=False):
        raise NotImplementedError

    def gist_delete(self, gist_id):
        raise NotImplementedError

    def request_create(self, user, repo, local_branch, remote_branch, title, description=None):
        raise NotImplementedError

    def request_list(self, user, repo):
        raise NotImplementedError

    def request_fetch(self, user, repo, request, pull=False):
        raise NotImplementedError

    def get_repository(self, user, repo):
        try:
            return self.gg.repository(user, repo)
        except ApiFailure as err:
            if err.status_code == 404:
                raise ResourceNotFoundError("Cannot get: repository {}/{} does not exists.".format(user, repo)) from err
            raise ResourceError("Unhandled error: {}".format(err)) from err
        except Exception as err:
            raise ResourceError("Unhandled exception: {}".format(err)) from err

    def get_parent_project_url(self, user, repo, rw=True): #pragma: no cover
        project = self.get_repository(user, repo)
        if not project.fork:
            return None
        if not hasattr(project, 'parent'):
            # Not yet in gogs_client
            # cf https://github.com/unfoldingWord-dev/python-gogs-client/pull/12
            log.warning('This project has an upstream, but the API does not give information on which.')
            return None
        elif project.parent:
            namespace, project = parent.full_name.split('/')
            return self.format_path(project, namespace, rw=rw)
        else:
            return None

    @staticmethod
    def get_project_default_branch(project):
        # not yet in gogs_client
        # cf https://github.com/unfoldingWord-dev/python-gogs-client/pull/8
        if hasattr(project, 'default_branch'):
            return project.default_branch
        return 'master'

    @staticmethod
    def is_repository_empty(project):
        # not yet in gogs_client
        # cf https://github.com/unfoldingWord-dev/python-gogs-client/pull/9
        if hasattr(project, 'empty'):
            return project.empty
        return False



#!/usr/bin/env python3

import os
import sys
import git

def extract_gitrepo_sections(conf):
    return filter(lambda k: k.startswith('gitrepo'), conf.sections())

def extract_gitrepo_conf(gconf_old, gconf_new):
    if os.path.exists(gconf_new):
        return "✗ cannot execute, file already exists: {}".format(gconf_new)
    with git.config.GitConfigParser(gconf_old, read_only=False) as cold:
        with git.config.GitConfigParser(gconf_new, read_only=False) as cnew:
            sections = list(extract_gitrepo_sections(cold))
            # copy the sections to the new configuration file
            cnew.update({s: {k:v for k,v in cold.items(s)} for s in sections})
            cnew.write()
            # remove the sections from the old configuration file
            for section in sections:
                cold.remove_section(section)
            # insert path to the new config file in the old one
            cold.update({'include': {'path': os.path.abspath(gconf_new)}})
    print("🍻 git-repo configuration extracted to new file: {}".format(gconf_new))

if __name__ == '__main__':
    if '-h' in sys.argv or '--help' in sys.argv:
        sys.exit('Usage: {} [.gitconfig-repos] [.gitconfig]'.format(sys.argv[0]))
    sys.exit(extract_gitrepo_conf(
        gconf_old=os.path.expanduser(len(sys.argv) >= 3 and sys.argv[2] or '~/.gitconfig'),
        gconf_new=os.path.expanduser(len(sys.argv) >= 2 and sys.argv[1] or '~/.gitconfig-repos')
    ))





#!/usr/bin/env python

import os
import sys
import logging

import pytest

#################################################################################
# Enable logging

log = logging.getLogger('test.gitbucket')

#################################################################################

from tests.helpers import GitRepoTestCase

from git_repo.services.service import gitbucket
from git_repo.exceptions import ResourceExistsError, ResourceNotFoundError, ResourceError


class Test_Gitbucket(GitRepoTestCase):
    log = log
    namespace = os.environ['GITBUCKET_NAMESPACE']


    @property
    def local_namespace(self):
        if 'GITBUCKET_NAMESPACE' in os.environ:
            return os.environ['GITBUCKET_NAMESPACE']
        return 'git-repo-user'

    def get_service(self):
        return gitbucket.GitbucketService(c={
            '__name__': 'gitrepo "gitbucket"',
            'fqdn': "localhost",
            'port': "8080",
            'scheme': "http",
            'insecure': 'yes',
            'ssh-url': 'ssh://git@localhost:29418'
            })

    def get_requests_session(self):
        return self.service.gh._session

    def test_00_fork(self):
        # AttributeError occurs, because GitBucket doesn't support repos/forks API.
        with pytest.raises(AttributeError):
            self.action_fork(local_namespace=self.local_namespace,
                             remote_namespace='root',
                             repository='repo')

    def test_01_create__new(self):
        self.action_create(namespace=self.local_namespace,
                           repository='foobar')

    def test_01_create__already_exists(self):
        # GitBucket doesn't return 422. it returns 200. maybe GitBucket's issue?
        # with pytest.raises(ResourceExistsError):
        self.action_create(namespace=self.local_namespace,
                        repository='git-repo')

    # skip because GitBucket doesn't support create group's repo by POST /api/v3/users/:org/repos
    @pytest.mark.skip
    def test_01_create_organization__new(self):
        self.action_create(namespace='group',
                           repository='foobar')

    # skip because GitBucket doesn't support create group's repo by POST /api/v3/users/:org/repos
    @pytest.mark.skip
    def test_01_create_organization__already_exists(self):
        with pytest.raises(ResourceExistsError):
            self.action_create(namespace='group',
                            repository='git-repo')

    def test_02_delete(self):
        with pytest.raises(NotImplementedError):
            self.action_delete(namespace=self.local_namespace,
                               repository='foobar')

    def test_03_delete_nouser(self):
        with pytest.raises(NotImplementedError):
            self.action_delete(repository='github3.py')

    def test_04_clone(self):
        self.action_clone(namespace='root',
                          repository='repo')

    def test_05_add(self):
        self.action_add(namespace='root',
                        repository='repo')

    def test_06_add__name(self):
        self.action_add(namespace='root',
                        repository='repo',
                        name='test0r')

    def test_07_add__alone(self):
        self.action_add(namespace='root',
                        repository='repo',
                        alone=True)

    def test_08_add__alone_name(self):
        self.action_add(namespace='root',
                        repository='repo',
                        name='test0r',
                        alone=True)

    def test_09_add__default(self):
        self.action_add(namespace='root',
                        repository='repo',
                        tracking='gitbucket')

    def test_10_add__default_name(self):
        self.action_add(namespace='root',
                        repository='repo',
                        name='test0r',
                        tracking='gitbucket')

    def test_11_add__alone_default(self):
        self.action_add(namespace='root',
                        repository='repo',
                        alone=True,
                        tracking='gitbucket')

    def test_12_add__alone_default_name(self):
        self.action_add(namespace='root',
                        repository='repo',
                        alone=True,
                        name='test0r',
                        tracking='gitbucket')

    def test_13_gist_list(self):
        with pytest.raises(NotImplementedError):
            g_list = []
            self.action_gist_list(gist_list_data=g_list)

    def test_14_gist_list_with_gist(self):
        with pytest.raises(NotImplementedError):
            g_list = [
                "C                 2542  i2c_scanner.c"
            ]
            self.action_gist_list(gist='10118958')

    def test_15_gist_list_with_bad_gist(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_list(gist='42')

    def test_16_gist_clone_with_gist(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_clone(gist='https://gist.github.com/10118958')

    def test_17_gist_fetch_with_gist(self):
        with pytest.raises(NotImplementedError):
            content = self.action_gist_fetch(gist='4170462', gist_file=None)
            assert content == '\n'.join([
                'diff --git a/platform/io.c b/platform/io.c',
                'index 209666a..0a6c2cf 100644',
                '--- a/platform/io.c',
                '+++ b/platform/io.c',
                '@@ -24,6 +24,16 @@',
                ' #if defined(__FreeBSD__)',
                ' #define IO_BSD',
                ' #elif defined(__APPLE__)',
                '+size_t strnlen(const char *s, size_t maxlen) ',
                '+{ ',
                '+        size_t len; ',
                '+ ',
                '+        for (len = 0; len < maxlen; len++, s++) { ',
                '+                if (!*s) ',
                '+                        break; ',
                '+        } ',
                '+        return (len); ',
                '+} ',
                ' #define IO_BSD',
                ' #define IO_USE_SELECT',
                ' #elif defined(WIN32)',
            ])

    def test_18_gist_fetch_with_bad_gist(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_fetch(gist='69', gist_file=None)

    def test_19_gist_fetch_with_gist_file(self):
        with pytest.raises(NotImplementedError):
            content = self.action_gist_fetch(gist='4170462', gist_file='freevpn0.029__platform__io.patch')
            assert content == '\n'.join([
                'diff --git a/platform/io.c b/platform/io.c',
                'index 209666a..0a6c2cf 100644',
                '--- a/platform/io.c',
                '+++ b/platform/io.c',
                '@@ -24,6 +24,16 @@',
                ' #if defined(__FreeBSD__)',
                ' #define IO_BSD',
                ' #elif defined(__APPLE__)',
                '+size_t strnlen(const char *s, size_t maxlen) ',
                '+{ ',
                '+        size_t len; ',
                '+ ',
                '+        for (len = 0; len < maxlen; len++, s++) { ',
                '+                if (!*s) ',
                '+                        break; ',
                '+        } ',
                '+        return (len); ',
                '+} ',
                ' #define IO_BSD',
                ' #define IO_USE_SELECT',
                ' #elif defined(WIN32)',
            ])

    def test_20_gist_fetch_with_bad_gist_file(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_fetch(gist='4170462', gist_file='failed')

    def test_21_gist_create_gist_file(self, datadir):
        with pytest.raises(NotImplementedError):
            test_file = str(datadir[ 'random-fortune-1.txt' ])
            self.action_gist_create(description='this is a test.',
                    gist_files=[ test_file ],
                    secret=False)

    def test_22_gist_create_gist_file_list(self, datadir):
        with pytest.raises(NotImplementedError):
            test_files = [
                    str(datadir[ 'random-fortune-1.txt' ]),
                    str(datadir[ 'random-fortune-2.txt' ]),
                    str(datadir[ 'random-fortune-3.txt' ]),
                    str(datadir[ 'random-fortune-4.txt' ]),
            ]
            self.action_gist_create(description='this is a test.',
                    gist_files=test_files,
                    secret=False)

    def test_23_gist_create_gist_dir(self, datadir):
        with pytest.raises(NotImplementedError):
            test_dir = [
                    str(datadir[ 'a_directory' ]),
            ]
            self.action_gist_create(description='this is a test.',
                    gist_files=test_dir,
                    secret=False)

    def test_24_gist_create_gist_file(self, datadir):
        with pytest.raises(NotImplementedError):
            test_file = str(datadir[ 'random-fortune-1.txt' ])
            self.action_gist_create(description='this is a secret test.',
                    gist_files=[ test_file ],
                    secret=True)

    def test_25_gist_create_gist_file_list(self, datadir):
        with pytest.raises(NotImplementedError):
            test_files = [
                    str(datadir[ 'random-fortune-1.txt' ]),
                    str(datadir[ 'random-fortune-2.txt' ]),
                    str(datadir[ 'random-fortune-3.txt' ]),
                    str(datadir[ 'random-fortune-4.txt' ]),
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_files,
                    secret=True)

    def test_26_gist_create_gist_dir(self, datadir):
        with pytest.raises(NotImplementedError):
            test_dir = [
                    str(datadir[ 'a_directory' ]),
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_dir,
                    secret=True)

    def test_27_gist_delete(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_delete(gist='7dcc495dda5e684cba94940a01f60e95')

    def test_28_gist_delete__not_exist(self):
        with pytest.raises(NotImplementedError):
            self.action_gist_delete(gist='7dcc495dda5e684cba94940a01f60e95')

    def test_29_gist_create_gist__file_not_exist(self, datadir):
        with pytest.raises(NotImplementedError):
            test_dir = [
                    'does_not_exists'
            ]
            self.action_gist_create(description='this is a secret test.',
                    gist_files=test_dir,
                    secret=False)

    @pytest.mark.skip
    def test_30_request_list(self):
        self.action_request_list(
                namespace='root',
                repository='repo',
                rq_list_data=[
            '{}\t{:<60}\t{}',
            ('id', 'title', 'URL'),
            ('3', 'docs for fqdn > url', 'https://github.com/guyzmo/git-repo/pull/3'),
            ('2', 'prefer gitrepo.<target>.token > privatekey, docs', 'https://github.com/guyzmo/git-repo/pull/2'),
        ])

    @pytest.mark.skip
    def test_31_request_fetch(self):
        self.action_request_fetch(namespace='root',
                repository='repo',
                request='1',
                remote_branch='pull',
                local_branch='requests/gitbucket')

    @pytest.mark.skip
    def test_31_request_fetch__bad_request(self):
        with pytest.raises(ResourceNotFoundError):
            self.action_request_fetch(namespace='guyzmo',
                repository='git-repo',
                request='1',
                remote_branch='pull',
                local_branch='requests/github',
                fail=True)

    @pytest.mark.skip
    def test_32_request_create(self):
        r = self.action_request_create(namespace=self.namespace,
                repository='test_create_requests',
                source_branch='pr-test',
                target_branch='master',
                title='PR test',
                description='PR description')
        assert r == {
                'local': 'pr-test',
                'ref': 1,
                'remote': 'master',
                'url': 'https://github.com/{}/test_create_requests/pull/1'.format(self.namespace),
                }

    @pytest.mark.skip
    def test_32_request_create__bad_branch(self):
        with pytest.raises(ResourceError):
            self.action_request_create(namespace=self.namespace,
                    repository='test_create_requests',
                    source_branch='does_not_exists',
                    target_branch='master',
                    title='PR test',
                    description='PR description')

    @pytest.mark.skip
    def test_32_request_create__bad_repo(self):
        with pytest.raises(ResourceNotFoundError):
            r = self.action_request_create(namespace=self.namespace,
                    repository='does_not_exists',
                    source_branch='pr-test',
                    target_branch='master',
                    title='PR test',
                    description='PR description')

    @pytest.mark.skip
    def test_32_request_create__guess_branch(self):
        r = self.action_request_create(namespace=self.namespace,
                repository='test_create_requests',
                source_branch=None,
                target_branch=None,
                title='PR test',
                description='PR description')

    def test_33_open(self):
        self.action_open(namespace='root',
                         repository='repo')

    @pytest.mark.skip
    def test_34_list__short(self, caplog):
        projects = self.action_list(namespace='group')
        assert projects == ['{}', ('Total repositories: 1',), ['git-repo-test/git-repo']]
        assert 'GET https://api.github.com/users/git-repo-test/repos' in caplog.text

    @pytest.mark.skip
    def test_34_list__long(self, caplog):
        projects = self.action_list(namespace='git-repo-test', _long=True)
        assert projects == ['{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{:12}\t{}',
                ['Status', 'Commits', 'Reqs', 'Issues', 'Forks', 'Coders', 'Watch', 'Likes', 'Lang', 'Modif', 'Name'],
                ['F ', '92', '0', '0', '0', '1', '0', '0', 'Python', 'Mar 30 2016', 'git-repo-test/git-repo']]
        assert 'GET https://api.github.com/users/git-repo-test/repos' in caplog.text





#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.gitbucket')

from ..service import register_target
from .github import GithubService
from ...exceptions import ArgumentError

import github3

@register_target('bucket', 'gitbucket')
class GitbucketService(GithubService):
    fqdn = "localhost"
    port = 8080

    def __init__(self, *args, **kwarg):
        super(GitbucketService, self).__init__(*args, **kwarg)

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        print("Please open the following URL: {}/{}/_application".format(cls.build_url(cls), login))
        print("Generate a new token, and paste it at the following prompt.")
        return prompt('token> ')

    def format_path(self, repository, namespace=None, rw=False):
        repo = repository
        if namespace:
            repo = '{}/{}'.format(namespace, repository)

        if not rw and '/' in repo:
            return '{}/git/{}.git'.format(self.url_ro, repo)
        elif rw and '/' in repo:
            if 'ssh://' in self.url_rw:
                return '{}/{}.git'.format(self.url_rw, repo)
            else:
                return '{}:{}.git'.format(self.url_rw, repo)
        else:
            raise ArgumentError("Cannot tell how to handle this url: `{}/{}`!".format(namespace, repo))

    def delete(self, repo, user=None):
        raise NotImplementedError("GitBucket doesn't suport this action now.")

    def request_create(self, user, repo, from_branch, onto_branch, title=None, description=None, auto_slug=False, edit=None):
        raise NotImplementedError("GitBucket doesn't support this action now.")

    def gist_list(self, gist=None):
        raise NotImplementedError("GitBucket doesn't support manipulate gist by API.")

    def gist_fetch(self, gist, fname=None):
        raise NotImplementedError("GitBucket doesn't support manipulate gist by API.")

    def gist_clone(self, gist):
        raise NotImplementedError("GitBucket doesn't support manipulate gist by API.")

    def gist_create(self, gist_pathes, description, secret=False):
        raise NotImplementedError("GitBucket doesn't support manipulate gist by API.")

    def gist_delete(self, gist_id):
        raise NotImplementedError("GitBucket doesn't support manipulate gist by API.")



#!/usr/bin/env python

import logging
import pytest

#################################################################################
# Enable logging

log = logging.getLogger('test.main')

#################################################################################

from tests.helpers import GitRepoMainTestCase

class Test_Main(GitRepoMainTestCase):
    log = log
    target = 'hub'

    def test_add(self, capsys):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': None,
                'alone': False,
                'auto_slug': False,
                'tracking': 'master'} == seen_args
        out, err = capsys.readouterr()
        assert 'Successfully added `guyzmo/git-repo` as remote named `foobar`\n' == err

    def test_add__alone(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'--alone': True})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': None,
                'alone': True,
                'auto_slug': False,
                'tracking': 'master'} == seen_args

    def test_add__tracking(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'--tracking': 'foobar'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': None,
                'alone': False,
                'auto_slug': False,
                'tracking': 'foobar'} == seen_args

    def test_add__tracking_alone(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'--alone': True,
                                                   '--tracking': 'foobar'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': None,
                'alone': True,
                'auto_slug': False,
                'tracking': 'foobar'} == seen_args

    def test_add__name(self, capsys):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'<name>': 'foobar'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': 'foobar',
                'alone': False,
                'auto_slug': False,
                'tracking': 'master'} == seen_args
        out, err = capsys.readouterr()
        assert 'Successfully added `guyzmo/git-repo` as remote named `foobar`\n' in err

    def test_add__name_tracking(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'<name>': 'foobar',
                                                   '--tracking': 'barfoo'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': 'foobar',
                'alone': False,
                'auto_slug': False,
                'tracking': 'barfoo'} == seen_args

    def test_add__name_alone(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'--alone': True,
                                                   '<name>': 'foobar'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': 'foobar',
                'alone': True,
                'auto_slug': False,
                'tracking': 'master'} == seen_args

    def test_add__name_alone_tracking(self):
        repo_slug, seen_args = self.main_add('guyzmo/git-repo', 0,
                                             args={'--alone': True,
                                                   '<name>': 'foobar',
                                                   '--tracking': 'barfoo'})
        assert ('guyzmo', 'git-repo', ) == repo_slug
        assert {'name': 'foobar',
                'alone': True,
                'auto_slug': False,
                'tracking': 'barfoo'} == seen_args

    def test_clone(self):
        repo_slug_branch, seen_args = self.main_clone('guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo', 'master') == repo_slug_branch

    def test_clone__branch(self):
        repo_slug_branch, seen_args = self.main_clone('guyzmo/git-repo', 0,
                                                      args={'<branch>': 'foobar'})
        assert ('guyzmo', 'git-repo', 'foobar') == repo_slug_branch

    def test_clone__full_url_http(self):
        repo_slug_branch, seen_args = self.main_clone('http://service.com/guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo', 'master') == repo_slug_branch

    def test_clone__full_url_https(self):
        repo_slug_branch, seen_args = self.main_clone('https://service.com/guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo', 'master') == repo_slug_branch

    def test_clone__full_url_git(self):
        repo_slug_branch, seen_args = self.main_clone('git@service.com/guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo', 'master') == repo_slug_branch

    def test_clone__too_many_slashes(self):
        did_clone = self.main_clone('guyzmo/git/repo', 2)
        assert None is did_clone

    def test_create(self):
        repo_slug, seen_args = self.main_create('guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {'add': False} == seen_args

    def test_create__add(self):
        repo_slug, seen_args = self.main_create('guyzmo/git-repo', 0, args={'--add': True})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {'add': True} == seen_args

    def test_create__no_such_path(self):
        did_add = self.main_add('guyzmo/git-repo', 2,
                      {'--path': '/'})
        assert did_add is None

    def test_delete(self):
        repo_slug, seen_args = self.main_delete('guyzmo/git-repo', 0, args={'--force': True})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_delete__ask(self):
        import io, sys
        stdin = sys.stdin
        sys.stdin = io.StringIO('y\nburn!')
        repo_slug, seen_args = self.main_delete('guyzmo/git-repo', 0)
        sys.stdin = stdin
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_delete__no_user(self):
        repo_slug, seen_args = self.main_delete('git-repo', 0, args={'--force': True})
        assert ('git-repo',) == repo_slug
        assert {} == seen_args

    def test_fork(self):
        repo_slug, seen_args = self.main_fork('guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_fork__branch(self):
        repo_slug, seen_args = self.main_fork('guyzmo/git-repo', 0,
                                              args={'<branch>': 'foobar'})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_fork__clone(self):
        repo_slug, seen_args = self.main_fork('guyzmo/git-repo', 0,
                                              args={'--clone': True})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_fork__branch_clone(self):
        repo_slug, seen_args = self.main_fork('guyzmo/git-repo', 0,
                                              args={'--clone': True,
                                                    '<branch>': 'foobar'})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_gist_list(self, capsys, caplog):
        did_list = self.main_gist_list(0, args={})
        out, err = capsys.readouterr()
        assert did_list == ((None,), {})
        assert out ==  ''.join([
            'id1 value1\n',
            'id2 value2\n',
            'id3 value3\n'
            ])

    def test_gist_ls(self, capsys, caplog):
        did_list = self.main_gist_ls(0, args={})
        out, err = capsys.readouterr()
        assert did_list == ((None,), {})
        assert out ==  ''.join([
            'id1 value1\n',
            'id2 value2\n',
            'id3 value3\n'
            ])

    def test_gist_list__with_gist(self, capsys, caplog):
        did_list = self.main_gist_list(0, args={'<gist>': 'foobar'})
        out, err = capsys.readouterr()
        assert did_list == (('foobar',), {})
        assert out ==  ''.join([
            'lang1 size1 name1\n',
            'lang2 size2 name2\n',
            'lang3 size3 name3\n'
            ])

    def test_gist_list__with_bad_gist(self, caplog):
        did_list = self.main_gist_list(2, args={'<gist>': 'bad', '--verbose': 0})
        assert did_list == (('bad',), {})
        assert 'Fatal error: bad gist!' in caplog.text

    def test_gist_clone__with_gist(self, caplog):
        did_clone = self.main_gist_clone(0, args={'<gist>': '42'})
        assert did_clone == (('42',), {})
        assert 'git init' in caplog.text
        assert 'Successfully cloned `42` into `{}/42`!'.format(self.tempdir.name) in caplog.text

    def test_gist_clone__with_bad_gist(self, caplog):
        did_clone = self.main_gist_clone(2, args={'<gist>': 'bad', '--verbose': 0})
        assert did_clone == (('bad',), {})
        assert 'Fatal error: bad gist!\n' in caplog.text

    def test_gist_fetch__with_gist(self, capsys):
        did_fetch = self.main_gist_fetch(0, args={'<gist>': '42'})
        out, err = capsys.readouterr()
        assert did_fetch == (('42', None), {})
        assert out == 'content of a gist\n'
        assert err == ''

    def test_gist_fetch__with_bad_gist(self, caplog):
        did_fetch = self.main_gist_fetch(2, args={'<gist>': 'bad', '--verbose': 0})
        assert did_fetch == (('bad', None), {})
        assert 'Fatal error: bad gist!\n' in caplog.text

    def test_gist_fetch__with_gist_file(self, capsys):
        did_fetch = self.main_gist_fetch(0, args={'<gist>': '42', '<gist_file>': 'fubar'})
        out, err = capsys.readouterr()
        assert did_fetch == (('42', 'fubar'), {})
        assert out == 'content of a gist\n'
        assert err == ''

    def test_gist_fetch__with_bad_gist_file(self, caplog):
        did_fetch = self.main_gist_fetch(2, args={'<gist>': '42', '<gist_file>': 'bad', '--verbose': 0})
        assert did_fetch == (('42', 'bad'), {})
        assert 'Fatal error: bad gist file!\n' in caplog.text

    def test_gist_create(self, caplog):
        did_create = self.main_gist_create(0, args={
            '<gist_path>': ['f1'],
            '<description>': 'fubar',
            '--secret': False,
            '--verbose': 0

        })
        assert did_create == ((['f1'], 'fubar', False), {})
        assert 'Successfully created gist `https://gists/42`!\n' in caplog.text

    def test_gist_create__exists(self, caplog):
        did_create = self.main_gist_create(2, args={
            '<gist_path>': ['exists'],
            '<description>': 'fubar',
            '--secret': False,
            '--verbose': 0
        })
        assert did_create == ((['exists'], 'fubar', False), {})
        assert 'Fatal error: gist exists!' in caplog.text

    def test_gist_create__secret(self, caplog):
        did_create = self.main_gist_create(0, args={
            '<gist_path>': ['f1'],
            '<description>': 'fubar',
            '--secret': True,
            '--verbose': 0

        })
        assert did_create == ((['f1'], 'fubar', True), {})
        assert 'Successfully created gist `https://gists/42`!\n' in caplog.text

    def test_gist_create__secret__exists(self, caplog):
        did_create = self.main_gist_create(2, args={
            '<gist_path>': ['exists'],
            '<description>': 'fubar',
            '--secret': True,
            '--verbose': 0
        })
        assert did_create == ((['exists'], 'fubar', True), {})
        assert 'Fatal error: gist exists!' in caplog.text

    def test_gist_delete__force(self, caplog):
        did_delete = self.main_gist_delete(0, args={'--force': True})
        assert 'Successfully deleted gist!' in caplog.text

    def test_gist_delete__ask(self, caplog):
        import io, sys
        stdin = sys.stdin
        sys.stdin = io.StringIO('y\nburn!')
        did_delete = self.main_gist_delete(0, args={'--force': True})
        sys.stdin = stdin
        assert 'Successfully deleted gist!' in caplog.text

    def test_request_list(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        repo_slug, seen_args = self.main_request_list('guyzmo/git-repo', 0, args={})
        out, err = capsys.readouterr()
        assert out ==  '1 desc1 http://request/1\n2 desc2 http://request/2\n3 desc3 http://request/3\n'

    def test_request_list__with_repo_slug__no_repo(self, capsys, caplog):
        repo_slug, seen_args = self.main_request_list('guyzmo/git-repo', 0, args={})
        out, err = capsys.readouterr()
        assert out ==  '1 desc1 http://request/1\n2 desc2 http://request/2\n3 desc3 http://request/3\n'

    # Commented out because this does not work on travis CI
    # def test_request_list__no_repo_slug__git(self, capsys, caplog):
    #     from subprocess import call
    #     call(['git', 'init', '-q', self.tempdir.name])
    #     call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/git-repo'])
    #     repo_slug, seen_args = self.main_request_list(rc=0, args={})
    #     out, err = capsys.readouterr()
    #     assert ('guyzmo', 'git-repo') == repo_slug
    #     assert dict() == seen_args
    #     assert out ==  '  1\tdesc1                                                       \thttp://request/1\n  2\tdesc2                                                       \thttp://request/2\n  3\tdesc3                                                       \thttp://request/3\n'
    #     assert 'id' in caplog.text and 'title' in caplog.text and 'URL' in caplog.text

    def test_request_list__no_repo_slug__https(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/git-repo'])
        repo_slug, seen_args = self.main_request_list(rc=0, args={})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert dict() == seen_args
        out, err = capsys.readouterr()
        assert out ==  '1 desc1 http://request/1\n2 desc2 http://request/2\n3 desc3 http://request/3\n'

    def test_request_list__no_repo_slug__https_dot_git_fix__issue55(self):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/.git-repo'])
        with self.mockup_git('guyzmo', '.git-repo', 'https://github.com/guyzmo/.git-repo'):
            self.set_mock_popen_commands([
                ('git remote get-url --all github', b'https://github.com/guyzmo/.git-repo\n', b'', 0),
            ])
            repo_slug, seen_args = self.main_request_list(rc=0, args={})
            assert ('guyzmo', '.git-repo') == repo_slug
            assert dict() == seen_args

    def test_request_list__no_repo_slug__git_dot_git_fix__issue55(self):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'git@github.com:guyzmo/.git-repo'])
        with self.mockup_git('guyzmo', '.git-repo', 'git@github.com:guyzmo/.git-repo'):
            self.set_mock_popen_commands([
                ('git remote get-url --all github', b'git@github.com:guyzmo/.git-repo', b'', 0),
            ])
            repo_slug, seen_args = self.main_request_list(rc=0, args={})
            assert ('guyzmo', '.git-repo') == repo_slug
            assert dict() == seen_args

    def test_request_fetch__request(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_fetch('guyzmo/git-repo', 0,
                args={'<request>': '42'})
        out, err = capsys.readouterr()
        assert ('guyzmo', 'git-repo', '42') == seen_args
        assert {'force': False} == extra_args
        assert out == ''
        assert 'Successfully fetched request id `42` of `guyzmo/git-repo` into `pr/42`!' in caplog.text

    # Commented out because this does not work on travis CI
    # def test_request_fetch__request__no_repo_slug__git(self, capsys, caplog):
    #     from subprocess import call
    #     call(['git', 'init', '-q', self.tempdir.name])
    #     call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/git-repo'])
    #     seen_args, extra_args = self.main_request_fetch(rc=0, args={'<request>': '42'})
    #     out, err = capsys.readouterr()
    #     assert ('guyzmo', 'git-repo', '42') == seen_args
    #     assert {'force': False} == extra_args
    #     assert out == ''
    #     assert 'Successfully fetched request id `42` of `guyzmo/git-repo` into `pr/42`!' in caplog.text

    def test_request_fetch__request__no_repo_slug__https(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        call(['git', '--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/git-repo'])
        seen_args, extra_args = self.main_request_fetch(rc=0, args={'<request>': '42'})
        out, err = capsys.readouterr()
        assert ('guyzmo', 'git-repo', '42') == seen_args
        assert {'force': False} == extra_args
        assert out == ''
        assert 'Successfully fetched request id `42` of `guyzmo/git-repo` into `pr/42`!' in caplog.text

    def test_request_fetch__bad_request(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_fetch('guyzmo/git-repo', 2,
                args={'<request>': 'bad', '--verbose': 0})
        out, err = capsys.readouterr()
        assert ('guyzmo', 'git-repo', 'bad') == seen_args
        assert {'force': False} == extra_args
        assert out == ''
        assert 'Fatal error: bad request for merge!' in caplog.text

    def test_request_create(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 0,
                args={
                    '<local_branch>': 'pr-test',
                    '<remote_branch>': 'base-test',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', 'pr-test', 'base-test', 'This is a test', 'This is a test', False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Successfully created request of `pr-test` onto `guyzmo/test:base-test`, with id `42`!' in caplog.text

    def test_request_create__no_description(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 0,
                args={
                    '<local_branch>': 'pr-test',
                    '<remote_branch>': 'base-test',
                    '<title>': 'This is a test',
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', 'pr-test', 'base-test', 'This is a test', None, False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Successfully created request of `pr-test` onto `guyzmo/test:base-test`, with id `42`!' in caplog.text

    def test_request_create__bad_local_branch(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 2,
                args={
                    '<local_branch>': 'bad',
                    '<remote_branch>': 'base-test',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', 'bad', 'base-test', 'This is a test', 'This is a test', False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Fatal error: bad branch to request!' in caplog.text

    def test_request_create__bad_remote_branch(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 2,
                args={
                    '<local_branch>': 'pr-test',
                    '<remote_branch>': 'bad',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', 'pr-test', 'bad', 'This is a test', 'This is a test', False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Fatal error: bad branch to request!' in caplog.text

    def test_request_create__no_local_branch(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 0,
                args={
                    '<remote_branch>': 'base-test',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', None, 'base-test', 'This is a test', 'This is a test', False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Successfully created request of `pr-test` onto `guyzmo/test:base-test`, with id `42`!' in caplog.text

    def test_request_create__no_remote_branch(self, capsys, caplog):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        seen_args, extra_args = self.main_request_create('guyzmo/test', 0,
                args={
                    '<local_branch>': 'pr-test',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'test', 'pr-test', None, 'This is a test', 'This is a test', False) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Successfully created request of `pr-test` onto `guyzmo/test:base-test`, with id `42`!' in caplog.text

    def test_open(self):
        repo_slug, seen_args = self.main_open('guyzmo/git-repo', 0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def _create_repository(self, ro=False):
        from subprocess import call
        call(['git', 'init', '-q', self.tempdir.name])
        if ro:
            call(['git','--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'https://github.com/guyzmo/git-repo'])
        else:
            call(['git','--git-dir={}/.git'.format(self.tempdir.name), 'remote', 'add', 'github', 'git@github.com:guyzmo/git-repo'])

    def test_open__no_repo_slug__https(self):
        self._create_repository(ro=True)
        repo_slug, seen_args = self.main_open(rc=0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    # Skipped because this does not work on travis CI
    @pytest.mark.skip
    def test_open__no_repo_slug__git(self):
        self._create_repository()
        repo_slug, seen_args = self.main_open(rc=0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_create__no_repo_slug(self):
        self._create_repository(ro=True)
        repo_slug, seen_args = self.main_create(rc=0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {'add': False} == seen_args

    def test_fork__no_repo_slug(self):
        self._create_repository(ro=True)
        repo_slug, seen_args = self.main_fork(rc=0)
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_delete__no_repo_slug(self):
        self._create_repository(ro=True)
        repo_slug, seen_args = self.main_delete(rc=0, args={'--force': True})
        assert ('guyzmo', 'git-repo') == repo_slug
        assert {} == seen_args

    def test_request_list__no_repo_slug(self, capsys, caplog):
        self._create_repository(ro=True)
        repo_slug, seen_args = self.main_request_list(rc=0, args={})
        out, err = capsys.readouterr()
        assert out ==  '1 desc1 http://request/1\n2 desc2 http://request/2\n3 desc3 http://request/3\n'

    def test_request_fetch__no_repo_slug(self, capsys, caplog):
        self._create_repository(ro=True)
        seen_args, extra_args = self.main_request_fetch(rc=0,
                args={'<request>': '42'})
        out, err = capsys.readouterr()
        assert ('guyzmo', 'git-repo', '42') == seen_args
        assert {'force': False} == extra_args
        assert out == ''
        assert 'Successfully fetched request id `42` of `guyzmo/git-repo` into `pr/42`!' in caplog.text

    def test_request_create__no_repo_slug(self, capsys, caplog):
        self._create_repository(ro=True)
        seen_args, extra_args = self.main_request_create(rc=0,
                args={
                    '<local_branch>': 'pr-test',
                    '<remote_branch>': 'base-test',
                    '<title>': 'This is a test',
                    '--message': 'This is a test'
                    })
        out, err = capsys.readouterr()
        seen_args = seen_args[:-1] # remove the passed edition function
        assert ('guyzmo', 'git-repo', 'pr-test', 'base-test', 'This is a test', 'This is a test', True) == seen_args
        assert {} == extra_args
        assert out == ''
        assert 'Successfully created request of `pr-test` onto `guyzmo/git-repo:base-test`, with id `42`!' in caplog.text

    def test_config(self, capsys, caplog):
        import sys, io, getpass
        getpass.getpass = input
        sys.stdin = io.StringIO('\n'.join(['y', 'n', 'user', 'pass', 'y', 'fubar', 'y']))
        #
        conf = self.main_config(target='hub', rc=0)
        assert sorted(['[gitrepo "github"]\n',
                '\tremote = fubar\n',
                '\ttoken = user:pass\n',
                '[alias]\n',
                '\ttest_command = repo test_command\n']) == sorted(conf)
        last_conf = conf

        sys.stdin = io.StringIO('\n'.join([
            'y', 'y', 'frama', 'framagit.org', '', '', '', '', 'user', 'pass', 'y', 'framagit', 'y']))
        #
        conf = self.main_config(target='lab', rc=0)
        assert sorted(last_conf+[
                '[gitrepo "frama"]\n',
                '\tfqdn = framagit.org\n',
                '\tremote = framagit\n',
                '\ttype = gitlab\n',
                '\tname = frama\n',
                '\tcommand = frama\n',
                '\tport = 443\n',
                '\tscheme = https\n',
                '\tinsecure = False\n',
                '\ttoken = user:pass\n', ]) == sorted(conf)
        last_conf = conf

        sys.stdin = io.StringIO('\n'.join([
            'y', 'y', 'test0r', 'localhost', '8080', 'n', 'user', 'pass', 'y', 'test', 'y']))
        #
        conf = self.main_config(target='lab', rc=0)
        assert sorted(last_conf+[
                '[gitrepo "test0r"]\n',
                '\tfqdn = localhost\n',
                '\tremote = test\n',
                '\ttype = gitlab\n',
                '\tname = test0r\n',
                '\tcommand = test0r\n',
                '\tport = 8080\n',
                '\tscheme = http\n',
                '\ttoken = user:pass\n', ]) == sorted(conf)
        last_conf = conf

        sys.stdin = io.StringIO('\n'.join([
            'y', 'y', 'selfsigned', 'server.local', '8443',
            '', '', 'y', '/etc/server.pem', 'user', 'pass', 'y', 'self', 'y']))
        #
        conf = self.main_config(target='lab', rc=0)
        assert sorted(last_conf+['[gitrepo "selfsigned"]\n',
                '\tfqdn = server.local\n',
                '\tremote = self\n',
                '\ttype = gitlab\n',
                '\tname = selfsigned\n',
                '\tcommand = selfsigned\n',
                '\tport = 8443\n',
                '\tscheme = https\n',
                '\tinsecure = False\n',
                '\tserver-cert = /etc/server.pem\n',
                '\ttoken = user:pass\n', ]) == sorted(conf)
        last_conf = conf

        sys.stdin = io.StringIO('\n'.join([
            'y', 'y', 'lamer', 'pwnme.com', '',
            '', 'y', 'user', 'pass', 'y', 'l', 'y']))
        #
        conf = self.main_config(target='lab', rc=0)
        assert sorted(last_conf+['[gitrepo "lamer"]\n',
                '\tfqdn = pwnme.com\n',
                '\tremote = l\n',
                '\ttype = gitlab\n',
                '\tname = lamer\n',
                '\tcommand = lamer\n',
                '\tport = 443\n',
                '\tscheme = https\n',
                '\tinsecure = True\n',
                '\ttoken = user:pass\n', ]) == sorted(conf)

    def test_z_noop(self):
        self.main_noop('guyzmo/git-repo', 1)




#!/usr/bin/env pytho

'''
Usage:
    {self} [--path=<path>] [-v...] <target> fork [--branch=<branch>]
    {self} [--path=<path>] [-v...] <target> create [--add]
    {self} [--path=<path>] [-v...] <target> delete [-f]
    {self} [--path=<path>] [-v...] <target> open
    {self} [--path=<path>] [-v...] <target> (list|ls) [-l] <user>
    {self} [--path=<path>] [-v...] <target> fork <user>/<repo> [--branch=<branch>]
    {self} [--path=<path>] [-v...] <target> fork <user>/<repo> <repo> [--branch=<branch>]
    {self} [--path=<path>] [-v...] <target> create <user>/<repo> [--add]
    {self} [--path=<path>] [-v...] <target> delete <user>/<repo> [-f]
    {self} [--path=<path>] [-v...] <target> open <user>/<repo>
    {self} [--path=<path>] [-v...] <target> clone <user>/<repo> [<repo> [<branch>]]
    {self} [--path=<path>] [-v...] <target> add
    {self} [--path=<path>] [-v...] <target> add <user>/<repo> [<name>] [--tracking=<branch>] [-a]
    {self} [--path=<path>] [-v...] <target> request (list|ls)
    {self} [--path=<path>] [-v...] <target> request fetch <request> [-f]
    {self} [--path=<path>] [-v...] <target> request create [--title=<title>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> request create <local_branch> [--title=<title>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> request create <remote_branch> <local_branch> [--title=<title>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> request <user>/<repo> (list|ls)
    {self} [--path=<path>] [-v...] <target> request <user>/<repo> fetch <request> [-f]
    {self} [--path=<path>] [-v...] <target> request <user>/<repo> create [--title=<title>] [--branch=<remote>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> request <user>/<repo> create <local_branch> [--title=<title>] [--branch=<remote>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> request <user>/<repo> create <remote_branch> <local_branch> [--title=<title>] [--branch=<remote>] [--message=<message>]
    {self} [--path=<path>] [-v...] <target> (gist|snippet) (list|ls) [<gist>]
    {self} [--path=<path>] [-v...] <target> (gist|snippet) clone <gist>
    {self} [--path=<path>] [-v...] <target> (gist|snippet) fetch <gist> [<gist_file>]
    {self} [--path=<path>] [-v...] <target> (gist|snippet) create [--secret] <description> [<gist_path> <gist_path>...]
    {self} [--path=<path>] [-v...] <target> (gist|snippet) delete <gist> [-f]
    {self} [--path=<path>] [-v...] <target> config [--config=<gitconfig>]
    {self} [-v...] config [--config=<gitconfig>]
    {self} --help

Tool for managing remote repository services.

Commands:
    add                      Add the service as a remote on this repository
    clone                    Clones this repository from the service
    fork                     Fork (and clone) the repository from the service
    create                   Make this repository a new remote on the service
    delete                   Delete the remote repository
    list                     Lists the repositories for a given user
    gist                     Manages gist files
    request                  Handles requests for merge
    open                     Open the given or current repository in a browser
    config                   Run authentication process and configure the tool

Options:
    <user>/<repo>            Repository to work with
    -p,--path=<path>         Path to work on [default: .]
    -v,--verbose             Makes it more chatty (repeat twice to see git commands)
    -h,--help                Shows this message

Options for list:
    <user>                   Name of the user whose repositories will be listed
    -l,--long                Show one repository per line, when set show the results
                             with the following columns:
    STATUS, COMMITS, REQUESTS, ISSUES, FORKS, CONTRIBUTORS, WATCHERS, LIKES, LANGUAGE, MODIF, NAME

Options for add:
    <name>                   Name to use for the remote (defaults to name of repo)
    -t,--tracking=<branch>   Makes this remote tracking for the current branch
    -a,--alone               Does not add the remote to the 'all' remote

Option for both clone and fork:
    <repo>                   Name of the local workspace directory

Options for clone:
    <branch>                 Branch to pull [default: master]

Options for fork:
    --branch=<branch>        Branch to pull [default: master]

Options for create:
    --add                    Add to local repository after creation

Options for delete:
    -f,--force               Do not ask for confirmation

Options for gist:
    <gist>                   Identifier of the gist to fetch
    <gist_file>              Name of the file to fetch
    <gist_path>              Name of the file or directory to use for a new gist.
                             If path is a directory, all files directly within it
                             will be pushed. If a list of path is given, all files
                             from them will be pushed.
    --secret                 Do not publicize gist when pushing

Options for request:
    -t,--title=<title>       Title to give to the request for merge
    -m,--message=<message>   Description for the request for merge

Configuration options:
    alias                    Name to use for the git remote
    fqdn                     URL of the repository
    type                     Name of the service to use (github, gitlab, bitbucket)

Configuration example:

[gitrepo "gitlab"]
    token = yourapitoken
    alias = lab

[gitrepo "personal"]
    type = gitlab
    token = yourapitoken
    fqdn = custom.org

{self} version {version}, Copyright ©2016 Bernard `Guyzmo` Pratz
{self} comes with ABSOLUTELY NO WARRANTY; for more informations
read the LICENSE file available in the sources, or check
out: http://www.gnu.org/licenses/gpl-2.0.txt
'''

from docopt import docopt

import os
import sys
import json
import shutil
import logging
import pkg_resources

__version__ = pkg_resources.require('git-repo')[0].version
__author__ = 'Bernard `Guyzmo` Pratz <guyzmo+git_repo@m0g.net>'
__contributors__ = []

log_root = logging.getLogger()
log = logging.getLogger('git_repo')

if sys.version_info.major < 3: # pragma: no cover
    print('Please use with python version 3')
    sys.exit(1)

from .exceptions import ArgumentError, ResourceNotFoundError
from .services.service import RepositoryService

from .tools import print_tty, print_iter, loop_input, confirm
from .kwargparse import KeywordArgumentParser, store_parameter, register_action

from git import Repo, Git
from git.exc import InvalidGitRepositoryError, NoSuchPathError, BadName

import re

EXTRACT_URL_RE = re.compile('[^:]*(://|@)[^/]*/')


class GitRepoRunner(KeywordArgumentParser):

    def init(self): # pragma: no cover
        if 'GIT_WORK_TREE' in os.environ.keys() or 'GIT_DIR' in os.environ.keys():
            del os.environ['GIT_WORK_TREE']

    def get_service(self, lookup_repository=True, resolve_targets=None):
        if not lookup_repository:
            service = RepositoryService.get_service(None, self.target)
            service.connect()
        else:
            # Try to resolve existing repository path
            try:
                try:
                    repository = Repo(os.path.join(self.path, self.repo_name or ''))
                except NoSuchPathError:
                    repository = Repo(self.path)
            except InvalidGitRepositoryError:
                raise FileNotFoundError('Cannot find path to the repository.')
            service = RepositoryService.get_service(repository, self.target)
            if not self.repo_name:
                repo_slug = RepositoryService.guess_repo_slug(
                        repository, service, resolve_targets
                        )
                if repo_slug:
                    self.set_repo_slug(repo_slug, auto=True)
        return service

    '''Argument storage'''

    @store_parameter('--verbose')
    def set_verbosity(self, verbose): # pragma: no cover
        if verbose >= 5:
            print(self.args)
        if verbose >= 4:
            import http.client
            http.client.HTTPConnection.debuglevel = 1
            logging.getLogger("requests.packages.urllib3").setLevel(logging.DEBUG)
            logging.getLogger("requests.packages.urllib3").propagate = True
        if verbose >= 3:
            Git.GIT_PYTHON_TRACE = 'full'
        if verbose >= 2:
            Git.GIT_PYTHON_TRACE = True
            FORMAT = '> %(message)s'
            formatter = logging.Formatter(fmt=FORMAT)
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logging.getLogger('git.cmd').removeHandler(logging.NullHandler())
            logging.getLogger('git.cmd').addHandler(handler)
        if verbose >= 1:
            log_root.setLevel(logging.DEBUG)
        else:
            log_root.setLevel(logging.INFO)

        log.addHandler(logging.StreamHandler())

    @store_parameter('<user>/<repo>')
    def set_repo_slug(self, repo_slug, auto=False):
        self.repo_slug = EXTRACT_URL_RE.sub('', repo_slug) if repo_slug else repo_slug
        self._auto_slug = auto
        if not self.repo_slug:
            self.user_name = None
            self.repo_name = None
        elif '/' in self.repo_slug:
            # in case a full URL is given as parameter, just extract the slug part.
            self.user_name, self.repo_name, *overflow = self.repo_slug.split('/')
            if len(overflow) != 0:
                raise ArgumentError('Too many slashes.'
                                    'Format of the parameter is <user>/<repo> or <repo>.')
        else:
            self.user_name = None
            self.repo_name = self.repo_slug

    @store_parameter('<branch>')
    @store_parameter('--branch')
    def set_branch(self, branch):
        # FIXME workaround for default value that is not correctly parsed in docopt
        if branch == None:
            branch = 'master'

        self.branch = branch

    @store_parameter('<repo>')
    def set_target_repo(self, repo):
        self.target_repo = repo

    @store_parameter('<name>')
    def set_name(self, name):
        self.remote_name = name

    @store_parameter('<gist>')
    def set_gist_ref(self, gist):
        self.gist_ref = gist

    @store_parameter('--config')
    def store_gitconfig(self, val):
        self.config = val or RepositoryService.get_config_path()

    '''Actions'''

    @register_action('ls')
    @register_action('list')
    def do_list(self):
        print_iter(self.get_service(False).list(self.user, self.long))
        return 0

    @register_action('add')
    def do_remote_add(self):
        service = self.get_service()
        remote, user, repo = service.add(self.repo_name, self.user_name,
                    name=self.remote_name,
                    tracking=self.tracking,
                    alone=self.alone,
                    auto_slug=self._auto_slug)
        log.info('Successfully added `{}/{}` as remote named `{}`'.format(
            user, repo,
            remote.name)
        )
        return 0

    @register_action('fork')
    def do_fork(self):
        if not self.repo_slug:
            service = self.get_service(lookup_repository=True)
            try:
                repo_path = self.path
                repository = Repo(repo_path)
            except (InvalidGitRepositoryError, NoSuchPathError):
                raise ArgumentError('Path {} is not a git repository'.format(self.path))

        else:
            # git <target> fork <user>/<repo>
            if not self.target_repo:
                if not self.user_name:
                    raise ArgumentError('Cannot clone repository, '
                                        'you shall provide either a <user>/<repo> parameter '
                                        'or no parameters to fork current repository!')
                service = self.get_service(None)

            # git <target> fork <user>/<repo> <path>
            else:
                repo_path = os.path.join(self.path, self.target_repo)
                try:
                    service = RepositoryService.get_service(Repo(repo_path), self.target)
                except (InvalidGitRepositoryError, NoSuchPathError):
                    service = self.get_service(lookup_repository=False)
                    # if the repository does not exists at given path, clone upstream into that path
                    self.do_clone(service, repo_path)

        service.run_fork(self.user_name, self.repo_name, branch=self.branch)

        if not self.repo_slug or self.target_repo:
            log.info('Successfully forked {} as {} within {}.'.format(
                self.repo_slug, '/'.join([service.username, self.repo_name]), repo_path))

        return 0

    @register_action('clone')
    def do_clone(self, service=None, repo_path=None):
        service = service or self.get_service(lookup_repository=False)
        repo_path = repo_path or os.path.join(self.path, self.target_repo or self.repo_name)
        if os.path.exists(repo_path) and os.listdir(repo_path) != []:
            raise FileExistsError('Cannot clone repository, '
                                  'a folder named {} already exists and '
                                  'is not an empty directory!'.format(repo_path))
        try:
            repository = Repo.init(repo_path)
            service = RepositoryService.get_service(repository, self.target)
            service.clone(self.user_name, self.repo_name, self.branch)
            log.info('Successfully cloned `{}` into `{}`!'.format(
                service.format_path(self.repo_slug),
                repo_path)
            )
            return 0
        except Exception as err:
            if os.path.exists(repo_path):
                shutil.rmtree(repo_path)
            raise ResourceNotFoundError(err.args[2].decode('utf-8')) from err

    @register_action('create')
    def do_create(self):
        service = self.get_service(lookup_repository=self.repo_slug == None or self.add)
        # if no repo_slug has been given, use the directory name as current project name
        if not self.user_name and not self.repo_name:
            self.set_repo_slug('/'.join([service.user,
                os.path.basename(os.path.abspath(self.path))]))
        if not self.user_name:
            self.user_name = service.user
        service.create(self.user_name, self.repo_name, add=self.add)
        log.info('Successfully created remote repository `{}`, '
                 'with local remote `{}`'.format(
            service.format_path(self.repo_name, namespace=self.user_name),
            service.name)
        )
        return 0

    @register_action('delete')
    def do_delete(self):
        service = self.get_service(lookup_repository=self.repo_slug == None)
        if not self.force: # pragma: no cover
            if not confirm('repository', self.repo_slug):
                return 0

        if self.user_name:
            service.delete(self.repo_name, self.user_name)
        else:
            service.delete(self.repo_name)
        log.info('Successfully deleted remote `{}` from {}'.format(
            self.repo_slug,
            service.name)
        )
        return 0


    @register_action('open')
    def do_open(self):
        self.get_service(lookup_repository=self.repo_slug is None).open(self.user_name, self.repo_name)
        return 0

    @register_action('request', 'ls')
    @register_action('request', 'list')
    def do_request_list(self):
        service = self.get_service(lookup_repository=self.repo_slug == None)
        print_tty('List of open requests to merge:')
        print_iter(service.request_list(self.user_name, self.repo_name))
        return 0

    @register_action('request', 'create')
    def do_request_create(self):
        def request_edition(repository, from_branch, onto_target):
            try:
                commit = repository.commit(from_branch)
                title, *body = commit.message.split('\n')
            except BadName:
                log.error('Couldn\'t find local source branch {}'.format(from_branch))
                return None, None
            from tempfile import NamedTemporaryFile
            from subprocess import call
            with NamedTemporaryFile(
                    prefix='git-repo-issue-',
                    suffix='.md',
                    mode='w+b') as request_file:
                request_file.write((
                    '# Request for Merge Title ##########################\n'
                    '{}\n'
                    '\n'
                    '# Request for Merge Body ###########################\n'
                    '{}\n'
                    '####################################################\n'
                    '## Filled with commit:\n'
                    '## {}\n'
                    '####################################################\n'
                    '## To be applied:\n'
                    '##   from branch: {}\n'
                    '##   onto project: {}\n'
                    '####################################################\n'
                    '## * All lines starting with # will be ignored.\n'
                    '## * First non-ignored line is the title of the request.\n'
                      ).format(title, '\n'.join(body), commit.name_rev, from_branch, onto_target).encode('utf-8'))
                request_file.flush()
                rv = call("{} {}".format(os.environ['EDITOR'], request_file.name), shell=True)
                if rv != 0:
                    raise ArgumentError("Aborting request, as editor exited abnormally.")
                request_file.seek(0)
                request_message = map(lambda l: l.decode('utf-8'),
                        filter(lambda l: not l.strip().startswith(b'#'), request_file.readlines()))
                try:
                    title = next(request_message)
                    body = ''.join(request_message)
                except Exception:
                    raise ResourceError("Format of the request message cannot be parsed.")

                return title, body


        service = self.get_service(resolve_targets=('upstream', '{service}', 'origin'))

        new_request = service.request_create(self.user_name,
                self.repo_name,
                self.local_branch,
                self.remote_branch,
                self.title,
                self.message,
                self._auto_slug,
                request_edition)
        log.info('Successfully created request of `{local}` onto `{project}:{remote}`, with id `{ref}`!'.format(**new_request))
        if 'url' in new_request:
            log.info('available at: {url}'.format(**new_request))
        return 0

    @register_action('request', 'fetch')
    def do_request_fetch(self):
        service = self.get_service()
        new_branch = service.request_fetch(self.user_name, self.repo_name, self.request, force=self.force)
        log.info('Successfully fetched request id `{}` of `{}` into `{}`!'.format(
            self.request,
            self.repo_slug,
            new_branch)
        )
        return 0

    @register_action('gist', 'ls')
    @register_action('gist', 'list')
    @register_action('snippet', 'ls')
    @register_action('snippet', 'list')
    def do_gist_list(self):
        service = self.get_service(lookup_repository=False)
        print_iter(service.gist_list(self.gist_ref or None))
        return 0

    @register_action('gist', 'clone')
    @register_action('snippet', 'clone')
    def do_gist_clone(self):
        service = self.get_service(lookup_repository=False)
        repo_path = os.path.join(self.path, self.gist_ref.split('/')[-1])
        service.repository = Repo.init(repo_path)
        service.gist_clone(self.gist_ref)
        log.info('Successfully cloned `{}` into `{}`!'.format( self.gist_ref, repo_path))
        return 0

    @register_action('gist', 'fetch')
    @register_action('snippet', 'fetch')
    def do_gist_fetch(self):
        service = self.get_service(lookup_repository=False)
        # send gist to stdout, not using log.info on purpose here!
        print(service.gist_fetch(self.gist_ref, self.gist_file))
        return 0

    @register_action('gist', 'create')
    @register_action('snippet', 'create')
    def do_gist_create(self):
        service = self.get_service(lookup_repository=False)
        url = service.gist_create(self.gist_path, self.description, self.secret)
        log.info('Successfully created gist `{}`!'.format(url))
        return 0

    @register_action('gist', 'delete')
    @register_action('snippet', 'delete')
    def do_gist_delete(self):
        service = self.get_service(lookup_repository=False)
        if not self.force: # pragma: no cover
            if not confirm('snippet', self.gist_ref):
                return 0

        service.gist_delete(self.gist_ref)
        log.info('Successfully deleted gist!')
        return 0

    @register_action('config')
    def do_config(self):
        from getpass import getpass

        def setup_service(service):
            new_conf = dict(
                    fqdn=None,
                    remote=None,
                    )

            print('Is your service self-hosted?')
            if 'y' in input('    [yN]> ').lower():
                new_conf['type'] = service.name
                print('What name do you want to give this service?')
                new_conf['name'] = input('[{}]> '.format(service.name))
                new_conf['command'] = new_conf['name']
                service.name, service.command = new_conf['name'], new_conf['command']
                print('Enter the service\'s domain name:')
                new_conf['fqdn'] = input('[{}]> '.format(service.fqdn))
                print('Enter the service\'s port:')
                new_conf['port'] = input('[443]> ') or '443'
                print('Are you connecting using HTTPS? (you should):')
                if 'n' in input('    [Yn]> ').lower():
                    new_conf['scheme'] = 'http'
                else:
                    new_conf['scheme'] = 'https'
                    print('Do you need to use an insecure connection? (you shouldn\'t):')
                    new_conf['insecure'] = 'y' in input('    [yN]> ').lower()
                    service.session_insecure = new_conf['insecure']
                    if not new_conf['insecure']:
                        print('Do you want to setup the path to custom certificate?:')
                        if 'y' in input('    [yN]> ').lower():
                            new_conf['server-cert'] = loop_input('/path/to/certbundle.pem []> ')
                            service.session_certificate = new_conf['server-cert']

                service.fqdn = new_conf['fqdn']
                service.port = new_conf['port']
                service.scheme = new_conf['scheme']

            conf = service.get_config(self.config)
            if 'token' in conf:
                raise Exception('A token has been generated for this service. Please revoke and delete before proceeding.')

            print('Please enter your credentials to connect to the service:')
            username = loop_input('username> ')
            password = loop_input('password> ', method=getpass)

            new_conf['token'] = service.get_auth_token(username, password, prompt=loop_input)
            print('Great! You\'ve been identified 🍻')

            print('Do you want to give a custom name for this service\'s remote?')
            if 'y' in input('    [yN]> ').lower():
                print('Enter the remote\'s name:')
                new_conf['remote'] = loop_input('[{}]> '.format(service.name))

            print('Do you want to configure a git alias?')
            print('N.B.: instead of typing `git repo {0}` you\'ll be able to type `git {0}`'.format(service.command))
            if 'n' in input('    [Yn]> ').lower():
                set_alias = False
            else:
                set_alias = True

            service.store_config(self.config, **new_conf)
            if set_alias:
                service.set_alias(self.config)

        if self.target:
            services = [self.get_service(lookup_repository=False)]
        else:
            services = RepositoryService.service_map.values()

        for service in sorted(services, key=lambda s: s.name):
            print('Do you want to configure the {} service?'.format(service.name))
            if 'n' in input('    [Yn]> ').lower():
                continue
            setup_service(service)

        print('🍻 The configuration is done!')
        return 0


def main(args):
    try:
        return GitRepoRunner(args).run()
    except Exception as err:
        log.error('Fatal error: {}'.format(err))
        if log_root.level == logging.DEBUG:
            log.exception('------------------------------------')
        return 2

def cli(): #pragma: no cover
    try:
        sys.exit(main(docopt(__doc__.format(self=sys.argv[0].split(os.path.sep)[-1], version=__version__))))
    finally:
        # Whatever happens, make sure that the cursor reappears with some ANSI voodoo
        if sys.stdout.isatty():
            sys.stdout.write('\033[?25h')

if __name__ == '__main__': #pragma: no cover
    cli()



#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.base')

import os
import sys

from git import RemoteProgress, config as git_config
from progress.bar import IncrementalBar as Bar

from urllib.parse import ParseResult
from subprocess import call

from ..exceptions import (
        ArgumentError,
        ResourceError,
        ResourceNotFoundError,
        ResourceExistsError
)

'''select open command'''

if 'darwin' in sys.platform: #pragma: no cover
    OPEN_COMMAND = 'open'
else: #pragma: no cover
    OPEN_COMMAND = 'xdg-open'


class ProgressBar(RemoteProgress): # pragma: no cover
    '''Nice looking progress bar for long running commands'''
    def setup(self, repo_name):
        self.bar = Bar(message='Pulling from {}'.format(repo_name), suffix='')

    def update(self, op_code, cur_count, max_count=100, message=''):
        #log.info("{}, {}, {}, {}".format(op_code, cur_count, max_count, message))
        max_count = int(max_count or 100)
        if max_count != self.bar.max:
            self.bar.max = max_count
        self.bar.goto(int(cur_count))


def register_target(repo_cmd, repo_service):
    """Decorator to register a class with an repo_service"""
    def decorate(klass):
        log.debug('Loading service module class: {}'.format(klass.__name__) )
        klass.command = repo_cmd
        klass.name = repo_service
        RepositoryService.service_map[repo_service] = klass
        RepositoryService.command_map[repo_cmd] = repo_service
        return klass
    return decorate


class RepositoryService:
    '''Base class for all repository services'''
    service_map = dict()
    command_map = dict()

    # this symbol is made available for testing purposes
    _current = None

    config_options = [
            'type', 'token', 'alias', 'fqdn', 'remote',
            'port', 'scheme', 'insecure', 'name', 'command',
            'server-cert'
            ]

    @staticmethod
    def get_config_path():
        home_dir = os.path.expanduser('~')
        try:
            from xdg.BaseDirectory import xdg_config_home
        except ImportError:
            xdg_config_home = os.path.join(home_dir, ".config")
        for config in [
                os.path.join(home_dir, '.gitconfig'),
                os.path.join(xdg_config_home, 'git', 'config'),
            ]:
            if os.path.exists(config):
                return config
        raise ResourceNotFoundError('User\'s Git configuration file not found!')

    @staticmethod
    def convert_url_into_slug(url):
        if url.endswith('.git'):
            url = url[:-4]
        # strip http://, https:// and ssh://
        if '://' in url:
            *_, user, name = url.split('/')
            return '/'.join([user, name])
        # scp-style URL
        elif '@' in url and ':' in url:
            return url.split(':')[-1]

    @classmethod
    def guess_repo_slug(cls, repository, service, resolve_targets=None):
        if resolve_targets:
            targets = [target.format(service=service.name) for target in resolve_targets]
        else:
            targets = (service.name, 'upstream', 'origin')
        for remote in repository.remotes:
            if remote.name in targets:
                return cls.convert_url_into_slug(remote.url)
        return None

    @classmethod
    def get_config(cls, config):
        out = {}
        with git_config.GitConfigParser(config, read_only=True) as config:
            section = 'gitrepo "{}"'.format(cls.name)
            if config.has_section(section):
                for option in cls.config_options:
                    if config.has_option(section, option):
                        out[option] = config.get(section, option)
        return out

    @classmethod
    def store_config(cls, config, **kwarg):
        with git_config.GitConfigParser(config, read_only=False) as config:
            section = 'gitrepo "{}"'.format(kwarg.get('name', cls.name))
            for option, value in kwarg.items():
                if option not in cls.config_options:
                    raise ArgumentError('Option {} is invalid and cannot be setup.'.format(option))
                if value != None:
                    config.set_value(section, option, value)

    @classmethod
    def set_alias(cls, config):
        with git_config.GitConfigParser(config, read_only=False) as config:
            config.set_value('alias', cls.command, 'repo {}'.format(cls.command))

    @classmethod
    def get_service(cls, repository, command):
        '''Accessor for a repository given a command

        :param repository: git-python repository instance
        :param command: aliased name of the service
        :return: instance for using the service
        '''
        if not repository:
            config = git_config.GitConfigParser(cls.get_config_path())
        else:
            config = repository.config_reader()
        target = cls.command_map.get(command, command)
        conf_section = list(filter(lambda n: 'gitrepo' in n and target in n, config.sections()))

        http_section = [config._sections[scheme] for scheme in ('http', 'https') if scheme in config.sections()]

        # check configuration constraints
        if len(conf_section) == 0:
            if not target:
                raise ValueError('Service {} unknown'.format(target))
            else:
                config = dict()
        elif len(conf_section) > 1:
            raise ValueError('Too many configurations for service {}'.format(target))
        # get configuration section as a dict
        else:
            config = config._sections[conf_section[0]]

        if target in cls.service_map:
            service = cls.service_map.get(target, cls)
            service.name = target
        else:
            if 'type' not in config:
                raise ValueError('Missing service type for custom service.')
            if config['type'] not in cls.service_map:
                raise ValueError('Service type {} does not exists.'.format(config['type']))
            service = cls.service_map.get(config['type'], cls)

        cls._current = service(repository, config, http_section)
        return cls._current

    @classmethod
    def get_auth_token(cls, login, password, prompt=None):
        raise NotImplementedError

    def load_configuration(self, c, hc=[]):
        CONFIG_TRUE=('on', 'true', 'yes', '1')
        # if there's a configuration file, update the names accordingly
        if c:
            name = ' '.join(c['__name__'].replace('"', '').split(' ')[1:])
            if name != self.name:
                if 'fqdn' not in c:
                    raise ValueError('Custom services SHALL have an URL setting.')
                self.fqdn = c['fqdn']
                self.name = name
        # if not in the configuration file, retrieve the private key from the
        # environment (useful for travis configuration), otherwise, make it None.
        # using "token" > "private_token" > "privatekey" in configuration file to avoid
        # confusion with the SSH keys (yes that happened).
        # NB: `git config` doesn't parse underscores in option names, token.
        self._privatekey = os.environ.get('PRIVATE_KEY_{}'.format(self.name.upper()),
                                          c.get('token',
                                                c.get('private_token',
                                                      c.get('privatekey', None))))
        self._username = os.environ.get('USERNAME_{}'.format(self.name.upper()),
                                          c.get('username', None))
        self._alias = c.get('alias', self.name)

        self.fqdn = c.get('fqdn', self.fqdn)
        self.scheme = c.get('scheme', 'https')
        self.port = c.get('port', None)

        self.default_create_private = c.get('default-create-private', 'n').lower() in CONFIG_TRUE
        self.ssh_url = c.get('ssh-url', self.fqdn)

        self.session_insecure = c.get('insecure', 'false').lower() in CONFIG_TRUE
        self.session_certificate = c.get('certificate', None)
        self.session_proxy = {cf['__name__']: cf['proxy'] for cf in hc if cf.get('proxy', None)}

    def __init__(self, r=None, c=None, hc=[]):
        '''
        :param r: git-python repository instance
        :param c: configuration data

        Build a repository service instance, store configuration and parameters
        And launch the connection to the service
        '''

        self.repository = r
        self.config = c

        self.load_configuration(c, hc)

        # if service has a repository configured, connect
        if r:
            self.connect()

    '''URL handling'''

    '''name of the git user to use for SSH remotes'''
    git_user = 'git'

    @staticmethod
    def build_url(obj):
        scheme = getattr(obj, 'scheme', 'https')
        port = getattr(obj, 'port', None)
        # skip useless port specification
        if (not port
            or scheme == 'https' and str(port) == '443'
            or scheme == 'http' and str(port) == '80'):
            netloc = obj.fqdn
        else:
            netloc = '{}:{}'.format(obj.fqdn, port)
        return ParseResult(scheme, netloc, *['']*4).geturl()

    @property
    def url_ro(self):
        '''Property that returns the HTTP URL of the service'''
        return self.build_url(self)

    @property
    def url_rw(self):
        url = self.ssh_url
        return url if '@' in url else '@'.join([self.git_user, url])

    def _convert_user_into_remote(self, username, exclude=['all']):
        # builds a ref with an username and a branch
        # this method parses the repository's remotes to find the url matching username
        # and containing the given branch and returns the corresponding ref
        remotes = {remote.name: list(remote.urls) for remote in self.repository.remotes}
        for name in (self.name, 'upstream') + tuple(remotes.keys()):
            if name in remotes and name not in exclude:
                for url in remotes[name]:
                    if self.fqdn in url and username == url.split('/')[-2].split(':')[-1]:
                        yield name

    def format_path(self, repository, namespace=None, rw=False):
        '''format the repository's URL

        :param repository: name of the repository
        :param namespace: namespace of the repository
        :param rw: return a git+ssh URL if true, an https URL otherwise
        :return: the full URI of the repository ready to use as remote

        if namespace is not given, repository is expected to be of format
        `<namespace>/<repository>`.
        '''
        repo = repository
        if namespace:
            repo = '{}/{}'.format(namespace, repository)

        if not rw and '/' in repo:
            return '{}/{}'.format(self.url_ro, repo)
        elif rw and '/' in repo:
            if self.url_rw.startswith('ssh://'):
                return '{}/{}'.format(self.url_rw, repo)
            else:
                return '{}:{}'.format(self.url_rw, repo)
        else:
            raise ArgumentError("Cannot tell how to handle this url: `{}/{}`!".format(namespace, repo))

    def pull(self, remote, branch=None):
        '''Pull a repository
        :param remote: git-remote instance
        :param branch: name of the branch to pull
        '''
        pb = ProgressBar()
        pb.setup(self.name)
        if branch:
            remote.pull(branch, progress=pb)
        else: # pragma: no cover
            remote.pull(progress=pb)
        print()

    def fetch(self, remote, remote_branch, local_branch, force=False):
        '''Pull a repository
        :param remote: git-remote instance
        :param branch: name of the branch to pull
        '''
        pb = ProgressBar()
        pb.setup(self.name)
        remote.fetch(':'.join([remote_branch, local_branch]),
                update_head_ok=True, force=force, progress=pb)
        print()

    def clone(self, user, repo, branch=None, rw=True):
        '''Clones a new repository

        :param user: namespace of the repository
        :param repo: name slug of the repository
        :Param branch: branch to pull as tracking

        This command is fairly simple, and pretty close to the real `git clone`
        command, except it does not take a full path, but just a namespace/slug
        path for a given service.
        '''
        log.info('Cloning {}…'.format(repo))

        project = self.get_repository(user, repo)
        if not branch:
            branch = self.get_project_default_branch(project)

        is_empty = self.is_repository_empty(project)
        if is_empty:
            self.repository.init()
        else:
            url = self.get_parent_project_url(user, repo, rw=rw)
            if url:
                parent_user, parent_project = self.convert_url_into_slug(url).split('/')
                self.add(user=parent_user, repo=parent_project, name='upstream', alone=True)

        remote, *_ = self.add(user=user, repo=repo, tracking=True, rw=rw)
        if not is_empty:
            self.pull(remote, branch)

    def add(self, repo, user=None, name=None, tracking=False, alone=False, rw=True, auto_slug=False):
        '''Adding repository as remote

        :param repo: Name slug of the repository to add
        :param name: Name of the remote when stored
        :param tracking: When set, makes this remote the tracking for master operations
        :param alone: When set, exclude this remote from the "all" remote

        This method creates a new remote within the current repository *repo*.
        It chooses the name of the repository based on the service's name as
        held by current instance of RepositoryService, or uses the *name*
        parameter.

        It also creates an *all* remote that contains all the remotes added by
        this tool, to make it possible to push to all remotes at once.
        '''
        # git vcs add upstream
        if repo == 'upstream':
            try:
                slug = self.convert_url_into_slug(self.repository.remote(self.name).url)
                if not slug:
                    raise ValueError()
            except ValueError:
                raise ResourceNotFoundError('No existing remote for {} to find an upstream of'.format(self.name))

            url = self.get_parent_project_url(*slug.split('/'))
            if not url:
                raise ResourceNotFoundError('No upstream on {} found for this project.'.format(self.name))
            slug = self.convert_url_into_slug(url)
            if not slug:
                raise ResourceNotFoundError('No upstream on {} found for this project.'.format(self.name))
            user, repo = slug.split('/')
            name = 'upstream'
            alone = True
            auto_slug = False

        # git vcs add
        if auto_slug and not name:
            remote_urls = set()
            for remote in self.repository.remotes:
                for url in remote.urls:
                    if self.fqdn in url:
                        remote_urls.add(self.convert_url_into_slug(url))
            remote_urls = list(remote_urls)
            if len(remote_urls) > 0:
                raise ResourceNotFoundError('Couldn\'t find a remote to '
                        '{service}. Please run git add {service} user/project'.format(service=self.name))
            for idx, url in enumerate(remote_urls, 1):
                print("[{:d}] {}".format(idx, url))
            try:
                remote_url = int(input('Please choose an slug to use as \'{}\' remote: '.format(self.name)))
            except ValueError:
                raise ArgumentError('Please enter a valid integer value.')
            if remote_url < 1 or remote_url > idx:
                raise ArgumentError('Please enter one of the suggested choices.')
            user, repo = remote_urls[remote_url-1].split('/')
            name = self.name

        if not name:
            name = self.name

        if not user:
            if '/' in repo:
                user, repo = repo.split('/')
            else:
                raise ArgumentError('Unable to parse repository {}, missing path separator.'.format(repo))

        # removing remote if it already exists
        # and extract all repository
        all_remote = None
        for r in self.repository.remotes:
            if r.name == name:
                self.repository.delete_remote(r)
            # find and stash the remote 'all'
            elif r.name == 'all':
                all_remote = r
        # update remote 'all'
        if not alone:
            # if remote all does not exists
            if not all_remote:
                self.repository.create_remote('all', self.format_path(repo, user, rw=rw))
            else:
                url = self.format_path(repo, user, rw=True)
                # check if url not already in remote all
                if url not in all_remote.urls:
                    all_remote.set_url(new_url=self.format_path(repo, user, rw=rw), add=True)

        # adding "self" as the tracking remote
        if tracking:
            remote = self.repository.create_remote(name, self.format_path(repo, user, rw=rw))
            # lookup tracking branch (usually master)
            for branch in self.repository.branches:
                if tracking == branch.name:
                    # set that branch as tracking
                    branch.set_tracking_branch(remote.refs[0])
                    break
        else:
            remote = self.repository.create_remote(name, self.format_path(repo, user, rw=rw))

        return remote, user, repo


    def run_fork(self, user, repo, branch):
        if user == self.user:
        # forking the repository on the service
            raise ResourceError("Cannot fork a project from yourself.")
        log.info("Forking repository {}/{}…".format(user, repo))
        # checking for an 'upstream' remote.
        if self.repository:
            upstream_remotes = list(filter(lambda x: x.name == 'upstream', self.repository.remotes))
            if len(upstream_remotes) != 0:
                raise ResourceExistsError('A remote named `upstream` already exists. Has this repo already been forked?')
        fork_name = self.fork(user, repo)
        # checking if a remote with the service's name already exists
        if self.repository:
            service_remotes = list(filter(lambda x: x.name == self.name, self.repository.remotes))
            if len(service_remotes) != 0:
                # if it does, rename it to upstream
                self.repository.create_remote('upstream', service_remotes[0].url)
                self.repository.delete_remote(service_remotes[0].name)
            else:
                # otherwise create an upstream remote with the source y
                self.add(user=user, repo=repo, name='upstream', alone=True)
            # add the service named repository
            remote = self.add(repo=repo, user=self.user, tracking=self.name)
        log.info("New forked repository available at {}/{}".format(self.url_ro,
                                                                   fork_name))

    def open(self, user=None, repo=None):
        '''Open the URL of a repository in the user's browser'''
        call([OPEN_COMMAND, self.format_path(repo, namespace=user, rw=False)])

    def connect(self): #pragma: no cover
        '''Brings up the connection to the remote service's API

        Meant to be overloaded by subclass
        '''
        raise NotImplementedError

    def delete(self, repo, user=None): #pragma: no cover
        '''Delete a remote repository on the service

        :param repo: name of the remote repository to delete

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def list(self, user, _long=False):
        '''List an user's repositories on the service

        :param user: name of the user
        :param _long: format of the listing

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def create(self, user, repo, add=False): #pragma: no cover
        '''Create a new remote repository on the service

        :param repo: name of the repository to create

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def fork(self, user, repo): #pragma: no covr
        '''Forks a new remote repository on the service
        and pulls commits from it

        :param repo: name of the repository to create

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def gist_list(self):
        '''Lists gists

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def gist_fetch(self, gist): #pragma: no cover
        '''Fetches a published gist

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def gist_clone(self, gist): #pragma: no cover
        '''Clones a gist

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def gist_create(self, gist_path, secret=False): #pragma: no cover
        '''Pushes a new gist

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def gist_delete(self, gist_path, secret=False): #pragma: no cover
        '''Deletes a new gist

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def request_list(self, user, repo): #pragma: no cover
        '''Lists all available request for merging code
        sent to the remote repository

        :param repo: name of the repository to create

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def request_fetch(self, user, repo, request, pull=False, force=False): #pragma: no cover
        '''Fetches given request as a branch, and switch if pull is true

        :param repo: name of the repository to create

        Meant to be implemented by subclasses
        '''
        raise NotImplementedError

    def request_create(self, user, repo, from_branch, onto_branch, title, description=None, auto_slug=False):
        raise NotImplementedError

    @property
    def get_parent_project_url(self, user, repo, rw=True): #pragma: no cover
        raise NotImplementedError

    @property
    def user(self): #pragma: no cover
        raise NotImplementedError

    @staticmethod
    def is_repository_empty(project):
        raise NotImplementedError

    @staticmethod
    def get_project_default_branch(project):
        raise NotImplementedError

'''
register all services by importing their modules, from the ext pagckage

they are registered using the `register_target()` decorator, and added
to the `RepositorService.service_map` dictionary, and is accessed by the
`main()` function using the `RepositoryService.get_service()` method.
'''

from .ext import *




#!/usr/bin/env python3
# encoding: utf-8
from setuptools import setup, find_packages

import sys, os

import pip

from setuptools import setup, find_packages, dist
from setuptools.command.test import test as TestCommand
from distutils.core import Command
from distutils.core import setup

if sys.version_info.major < 3:
    print('Please install with python version 3')
    sys.exit(1)

# Use buildout's path for eggs
def get_egg_cache_dir(self):
    egg_cache_dir = os.path.join(os.curdir, 'var', 'eggs')
    if not os.path.exists(egg_cache_dir):
        os.makedirs(egg_cache_dir, exist_ok=True)
    return egg_cache_dir
dist.Distribution.get_egg_cache_dir = get_egg_cache_dir


class DistClean(Command):
    description = 'Clean the repository from all buildout stuff'
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        import shutil, os
        from glob import glob
        path = os.path.split(__file__)[0]
        shutil.rmtree(os.path.join(path, 'var'), ignore_errors=True)
        shutil.rmtree(os.path.join(path, 'bin'), ignore_errors=True)
        shutil.rmtree(os.path.join(path, '.tox'), ignore_errors=True)
        for fname in glob('*.egg-info'):
            shutil.rmtree(os.path.join(path, fname), ignore_errors=True)
        shutil.rmtree(os.path.join(path, '.eggs'), ignore_errors=True)
        shutil.rmtree(os.path.join(path, 'build'), ignore_errors=True)
        shutil.rmtree(os.path.join(path, 'dist'), ignore_errors=True)
        shutil.rmtree(os.path.join(path, '.installed.cfg'), ignore_errors=True)
        print("Repository is now clean!")


class PyTest(TestCommand):
    user_options = [
            ('exitfirst', 'x', "exit instantly on first error or failed test."),
            ('last-failed', 'l', "rerun only the tests that failed at the last run"),
            ('verbose', 'v', "increase verbosity"),
            ('pdb', 'p', "run pdb upon failure"),
            ('pep8', '8', "perform some pep8 sanity checks on .py files"),
            ('flakes', 'f', "run flakes on .py files"),
            ('pytest-args=', 'a', "Extra arguments to pass into py.test"),
            ]
    default_options = [
        '--cov=calenvite', '--cov-report', 'term-missing',
        '--capture=sys', 'tests'
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = set()
        self.exitfirst = False
        self.last_failed = False
        self.verbose = 0
        self.pdb = False
        self.pep8 = False
        self.flakes = False

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        if isinstance(self.pytest_args, str):
            self.pytest_args = set(self.pytest_args.split(' '))
        if self.exitfirst:   self.pytest_args.add('-x')
        if self.pdb:   self.pytest_args.add('--pdb')
        if self.last_failed: self.pytest_args.add('--last-failed')
        if self.pep8:        self.pytest_args.add('--pep8')
        if self.flakes:      self.pytest_args.add('--flakes')
        self.pytest_args = list(self.pytest_args)
        if self.verbose:
            for v in range(self.verbose):
                self.pytest_args.append('-v')
        errno = pytest.main(self.pytest_args + self.default_options)
        sys.exit(errno)


class Buildout(Command):
    description = 'Running buildout on the project'
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        try:
            from zc.buildout.buildout import main
        except ImportError:
            print('Please install buildout!\n  pip install zc.buildout')
            sys.exit(1)
        errno = main(sys.argv[sys.argv.index('buildout')+1:])
        if errno == 0:
            print('Now you can run tests using: bin/py.test')
            print('Now you can test current code using: bin/git-repo')
            print('Thank you 🍻')
        sys.exit(errno)


requirements_links = []

def requirements(spec=None):
    spec = '{}{}.txt'.format('requirements',
            '-'+spec if spec else '')
    requires = []

    requirements = pip.req.parse_requirements(
        spec, session=pip.download.PipSession())

    for item in requirements:
        if getattr(item, 'link', None):
            requirements_links.append(str(item.link))
        if item.req:
            requires.append(str(item.req))

    return requires


setup(name='git-repo',
      description='Tool for managing repository services from your git CLI tool',
      classifiers=[
          # 'Development Status :: 2 - Pre-Alpha',
          # 'Development Status :: 3 - Alpha',
          'Development Status :: 4 - Beta',
          # 'Development Status :: 5 - Production/Stable',
          # 'Development Status :: 6 - Mature',
          # 'Development Status :: 7 - Inactive',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Topic :: Software Development',
          'Topic :: Software Development :: Version Control',
          'Topic :: Utilities',
      ],
      keywords='git',
      url='https://github.com/guyzmo/git-repo',
      author='Bernard `Guyzmo` Pratz',
      author_email='guyzmo+git-repo@m0g.net',
      setup_requires=[
          'setuptools_scm',
          'setuptools-markdown',
          'wheel>=0.25.0',
      ],
      long_description_markdown_filename='README.md',
      use_scm_version={'version_scheme':'guess-next-dev'},
      include_package_data = True,
      install_requires=requirements(),
      tests_require=requirements('test'),
      dependency_links=requirements_links,
      cmdclass={
          'buildout': Buildout,
          'dist_clean': DistClean,
          'test': PyTest,
      },
      entry_points="""
      # -*- Entry points: -*-
      [console_scripts]
      git-repo = git_repo.repo:cli
      """,
      license='GPLv2',
      packages=find_packages(exclude=['tests']),
      test_suite='pytest',
      zip_safe=False
      )



#!/usr/bin/env python

import sys
import shutil

def print_tty(*args, **kwarg):
    if sys.stdout.isatty():
        print(*args, **kwarg)

def print_iter(generator):
    fmt = next(generator)
    print_tty(fmt.format(*next(generator)))
    for item in generator:
        print(fmt.format(*item))

def loop_input(*args, method=input, **kwarg):
    out = ''
    while len(out) == 0:
        out = method(*args, **kwarg)
    return out

def confirm(what, where):
    '''
    Method to show a CLI based confirmation message, waiting for a yes/no answer.
    "what" and "where" are used to better define the message.
    '''
    ans = input('Are you sure you want to delete the '
                '{} {} from the service?\n[yN]> '.format(what, where))
    if 'y' in ans:
        ans = loop_input('Are you really sure? there\'s no coming back!\n'
                    '[type \'burn!\' to proceed]> ')
        if 'burn!' != ans:
            return False
    else:
        return False
    return True

def columnize(lines, indent=0, pad=2):
    term_width = shutil.get_terminal_size((80, 20)).columns
    # prints a list of items in a fashion similar to the dir command
    # borrowed from https://gist.github.com/critiqjo/2ca84db26daaeb1715e1
    n_lines = len(lines)
    if n_lines == 0:
        return
    col_width = max(len(line) for line in lines)
    n_cols = int((term_width + pad - indent)/(col_width + pad))
    n_cols = min(n_lines, max(1, n_cols))
    col_len = int(n_lines/n_cols) + (0 if n_lines % n_cols == 0 else 1)
    if (n_cols - 1) * col_len >= n_lines:
        n_cols -= 1
    cols = [lines[i*col_len : i*col_len + col_len] for i in range(n_cols)]
    rows = list(zip(*cols))
    rows_missed = zip(*[col[len(rows):] for col in cols[:-1]])
    rows.extend(rows_missed)
    for row in rows:
        yield [" "*indent + (" "*pad).join(line.ljust(col_width) for line in row)]
